"""Dispense command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union, Any
from typing_extensions import Literal

from pydantic import Field
from pydantic.json_schema import SkipJsonSchema

from ..state.update_types import StateUpdate, CLEAR
from .pipetting_common import (
    PipetteIdMixin,
    DispenseVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    OverpressureError,
    dispense_in_place,
    DEFAULT_CORRECTION_VOLUME,
)
from .movement_common import (
    LiquidHandlingWellLocationMixin,
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..resources import ModelUtils
    from ..state.state import StateView


DispenseCommandType = Literal["dispense"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class DispenseParams(
    PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin, LiquidHandlingWellLocationMixin
):
    """Payload required to dispense to a specific well."""

    pushOut: float | SkipJsonSchema[None] = Field(
        None,
        description="push the plunger a small amount farther than necessary for accurate low-volume dispensing",
        json_schema_extra=_remove_default,
    )


class DispenseResult(BaseLiquidHandlingResult, DestinationPositionResult):
    """Result data from the execution of a Dispense command."""

    pass


_ExecuteReturn = Union[
    SuccessData[DispenseResult],
    DefinedErrorData[OverpressureError] | DefinedErrorData[StallOrCollisionError],
]


class DispenseImplementation(AbstractCommandImpl[DispenseParams, _ExecuteReturn]):
    """Dispense command implementation."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._pipetting = pipetting
        self._model_utils = model_utils

    async def execute(self, params: DispenseParams) -> _ExecuteReturn:
        """Move to and dispense to the requested well."""
        well_location = params.wellLocation
        labware_id = params.labwareId
        well_name = params.wellName
        volume = params.volume

        # TODO(pbm, 10-15-24): call self._state_view.geometry.validate_dispense_volume_into_well()

        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=params.pipetteId,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            operation_volume=volume,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result
        dispense_result = await dispense_in_place(
            pipette_id=params.pipetteId,
            volume=volume,
            flow_rate=params.flowRate,
            push_out=params.pushOut,
            location_if_error={
                "retryLocation": (
                    move_result.public.position.x,
                    move_result.public.position.y,
                    move_result.public.position.z,
                )
            },
            pipetting=self._pipetting,
            model_utils=self._model_utils,
            correction_volume=params.correctionVolume or DEFAULT_CORRECTION_VOLUME,
        )

        if isinstance(dispense_result, DefinedErrorData):
            return DefinedErrorData(
                public=dispense_result.public,
                state_update=(
                    StateUpdate.reduce(
                        move_result.state_update, dispense_result.state_update
                    ).set_liquid_operated(
                        labware_id=labware_id,
                        well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                            labware_id, well_name, params.pipetteId
                        ),
                        volume_added=CLEAR,
                    )
                ),
                state_update_if_false_positive=StateUpdate.reduce(
                    move_result.state_update,
                    dispense_result.state_update_if_false_positive,
                ),
            )
        else:
            volume_added = (
                self._state_view.pipettes.get_liquid_dispensed_by_ejecting_volume(
                    pipette_id=params.pipetteId, volume=dispense_result.public.volume
                )
            )
            if volume_added is not None:
                volume_added *= self._state_view.geometry.get_nozzles_per_well(
                    labware_id, well_name, params.pipetteId
                )
            return SuccessData(
                public=DispenseResult(
                    volume=dispense_result.public.volume,
                    position=move_result.public.position,
                ),
                state_update=(
                    StateUpdate.reduce(
                        move_result.state_update, dispense_result.state_update
                    ).set_liquid_operated(
                        labware_id=labware_id,
                        well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                            labware_id, well_name, params.pipetteId
                        ),
                        volume_added=volume_added
                        if volume_added is not None
                        else CLEAR,
                    )
                ),
            )


class Dispense(
    BaseCommand[
        DispenseParams, DispenseResult, OverpressureError | StallOrCollisionError
    ]
):
    """Dispense command model."""

    commandType: DispenseCommandType = "dispense"
    params: DispenseParams
    result: Optional[DispenseResult] = None

    _ImplementationCls: Type[DispenseImplementation] = DispenseImplementation


class DispenseCreate(BaseCommandCreate[DispenseParams]):
    """Create dispense command request model."""

    commandType: DispenseCommandType = "dispense"
    params: DispenseParams

    _CommandCls: Type[Dispense] = Dispense



from __future__ import annotations
from typing import TYPE_CHECKING, List, Sequence, Union, overload


from .helpers import (
    stringify_location,
    stringify_disposal_location,
    stringify_well_list,
    listify,
)
from . import types as command_types

from opentrons.types import Location
from opentrons.protocol_api.disposal_locations import TrashBin, WasteChute
from opentrons.protocol_api._liquid import LiquidClass

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Well


def home(mount: str) -> command_types.HomeCommand:
    text = f"Homing pipette plunger on mount {mount}"
    return {"name": command_types.HOME, "payload": {"axis": mount, "text": text}}


def aspirate(
    instrument: InstrumentContext,
    volume: float,
    location: Location,
    flow_rate: float,
    rate: float,
) -> command_types.AspirateCommand:
    location_text = stringify_location(location)
    template = "Aspirating {volume} uL from {location} at {flow} uL/sec"
    text = template.format(volume=float(volume), location=location_text, flow=flow_rate)

    return {
        "name": command_types.ASPIRATE,
        "payload": {
            "instrument": instrument,
            "volume": volume,
            "location": location,
            "rate": rate,
            "text": text,
        },
    }


def dispense(
    instrument: InstrumentContext,
    volume: float,
    location: Location,
    flow_rate: float,
    rate: float,
) -> command_types.DispenseCommand:
    location_text = stringify_location(location)
    template = "Dispensing {volume} uL into {location} at {flow} uL/sec"
    text = template.format(volume=float(volume), location=location_text, flow=flow_rate)

    return {
        "name": command_types.DISPENSE,
        "payload": {
            "instrument": instrument,
            "volume": volume,
            "location": location,
            "rate": rate,
            "text": text,
        },
    }


def dispense_in_disposal_location(
    instrument: InstrumentContext,
    volume: float,
    location: Union[TrashBin, WasteChute],
    flow_rate: float,
    rate: float,
) -> command_types.DispenseInDisposalLocationCommand:
    location_text = stringify_disposal_location(location)
    text = f"Dispensing {float(volume)} uL into {location_text} at {flow_rate} uL/sec"

    return {
        "name": command_types.DISPENSE_IN_DISPOSAL_LOCATION,
        "payload": {
            "instrument": instrument,
            "volume": volume,
            "location": location,
            "rate": rate,
            "text": text,
        },
    }


def consolidate(
    instrument: InstrumentContext,
    volume: Union[float, List[float]],
    source: List[Union[Location, Well]],
    dest: Union[Location, Well],
) -> command_types.ConsolidateCommand:
    text = "Consolidating {volume} from {source} to {dest}".format(
        volume=transform_volumes(volume),
        source=stringify_location(source),
        dest=stringify_location(dest),
    )
    locations: List[Union[Location, Well]] = listify(source) + listify(dest)
    return {
        "name": command_types.CONSOLIDATE,
        "payload": {
            "instrument": instrument,
            "locations": locations,
            "volume": volume,
            "source": source,
            "dest": dest,
            "text": text,
        },
    }


def distribute(
    instrument: InstrumentContext,
    volume: Union[float, List[float]],
    source: Union[Location, Well],
    dest: List[Union[Location, Well]],
) -> command_types.DistributeCommand:
    text = "Distributing {volume} from {source} to {dest}".format(
        volume=transform_volumes(volume),
        source=stringify_location(source),
        dest=stringify_location(dest),
    )
    locations: List[Union[Location, Well]] = listify(source) + listify(dest)
    return {
        "name": command_types.DISTRIBUTE,
        "payload": {
            "instrument": instrument,
            "locations": locations,
            "volume": volume,
            "source": source,
            "dest": dest,
            "text": text,
        },
    }


def transfer(
    instrument: InstrumentContext,
    volume: Union[float, List[float]],
    source: List[Union[Location, Well]],
    dest: List[Union[Location, Well]],
) -> command_types.TransferCommand:
    text = "Transferring {volume} from {source} to {dest}".format(
        volume=transform_volumes(volume),
        source=stringify_location(source),
        dest=stringify_location(dest),
    )
    locations: List[Union[Location, Well]] = listify(source) + listify(dest)
    return {
        "name": command_types.TRANSFER,
        "payload": {
            "instrument": instrument,
            "locations": locations,
            "volume": volume,
            "source": source,
            "dest": dest,
            "text": text,
        },
    }


@overload
def transform_volumes(volumes: Union[float, int]) -> float:
    ...


@overload
def transform_volumes(volumes: List[float]) -> List[float]:
    ...


def transform_volumes(
    volumes: Union[float, int, List[float]]
) -> Union[float, List[float]]:
    if not isinstance(volumes, list):
        return float(volumes)
    else:
        return [float(vol) for vol in volumes]


def mix(
    instrument: InstrumentContext,
    repetitions: int,
    volume: float,
    location: Union[Well, Location, None],
) -> command_types.MixCommand:
    text = "Mixing {repetitions} times with a volume of {volume} ul".format(
        repetitions=repetitions, volume=float(volume)
    )
    return {
        "name": command_types.MIX,
        "payload": {
            "instrument": instrument,
            "location": location,
            "volume": volume,
            "repetitions": repetitions,
            "text": text,
        },
    }


def blow_out(
    instrument: InstrumentContext, location: Location
) -> command_types.BlowOutCommand:
    location_text = stringify_location(location)
    text = f"Blowing out at {location_text}"

    return {
        "name": command_types.BLOW_OUT,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def blow_out_in_disposal_location(
    instrument: InstrumentContext, location: Union[TrashBin, WasteChute]
) -> command_types.BlowOutInDisposalLocationCommand:
    location_text = stringify_disposal_location(location)
    text = f"Blowing out into {location_text}"

    return {
        "name": command_types.BLOW_OUT_IN_DISPOSAL_LOCATION,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def touch_tip(instrument: InstrumentContext) -> command_types.TouchTipCommand:
    text = "Touching tip"

    return {
        "name": command_types.TOUCH_TIP,
        "payload": {"instrument": instrument, "text": text},
    }


def air_gap(
    instrument: InstrumentContext,
    volume: float | None,
    height: float | None,
) -> command_types.AirGapCommand:
    text = (
        "Air gap"
        + (f" of {volume} uL" if volume is not None else "")
        + (f" at height {height}" if height is not None else "")
    )
    return {
        "name": command_types.AIR_GAP,
        "payload": {
            "instrument": instrument,
            "volume": volume,
            "height": height,
            "text": text,
        },
    }


def return_tip() -> command_types.ReturnTipCommand:
    text = "Returning tip"
    return {"name": command_types.RETURN_TIP, "payload": {"text": text}}


def pick_up_tip(
    instrument: InstrumentContext, location: Well
) -> command_types.PickUpTipCommand:
    location_text = stringify_location(location)
    text = f"Picking up tip from {location_text}"
    return {
        "name": command_types.PICK_UP_TIP,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def drop_tip(
    instrument: InstrumentContext, location: Well
) -> command_types.DropTipCommand:
    location_text = stringify_location(location)
    text = "Dropping tip into {location}".format(location=location_text)
    return {
        "name": command_types.DROP_TIP,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def drop_tip_in_disposal_location(
    instrument: InstrumentContext, location: Union[TrashBin, WasteChute]
) -> command_types.DropTipInDisposalLocationCommand:
    location_text = stringify_disposal_location(location)
    text = f"Dropping tip into {location_text}"
    return {
        "name": command_types.DROP_TIP_IN_DISPOSAL_LOCATION,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def move_to(
    instrument: InstrumentContext,
    location: Location,
) -> command_types.MoveToCommand:
    location_text = stringify_location(location)
    text = "Moving to {location}".format(location=location_text)
    return {
        "name": command_types.MOVE_TO,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def move_to_disposal_location(
    instrument: InstrumentContext,
    location: Union[TrashBin, WasteChute],
) -> command_types.MoveToDisposalLocationCommand:
    location_text = stringify_disposal_location(location)
    text = f"Moving to {location_text}"
    return {
        "name": command_types.MOVE_TO_DISPOSAL_LOCATION,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def transfer_with_liquid_class(
    instrument: InstrumentContext,
    liquid_class: LiquidClass,
    volume: float,
    source: Union[Well, Sequence[Well], Sequence[Sequence[Well]]],
    destination: Union[
        Well, Sequence[Well], Sequence[Sequence[Well]], TrashBin, WasteChute
    ],
) -> command_types.TransferWithLiquidClassCommand:
    if isinstance(destination, (TrashBin, WasteChute)):
        destination_text = stringify_disposal_location(destination)
    else:
        destination_text = stringify_well_list(destination)
    text = (
        "Transferring "
        + f"{volume} uL of {liquid_class.display_name} liquid class from "
        + f"{stringify_well_list(source)} to {destination_text}"
    )
    return {
        "name": command_types.TRANSFER_WITH_LIQUID_CLASS,
        "payload": {
            "instrument": instrument,
            "liquid_class": liquid_class,
            "volume": volume,
            "source": source,
            "destination": destination,
            "text": text,
        },
    }


def distribute_with_liquid_class(
    instrument: InstrumentContext,
    liquid_class: LiquidClass,
    volume: float,
    source: Union[Well, Sequence[Well], Sequence[Sequence[Well]]],
    destination: Union[Well, Sequence[Well], Sequence[Sequence[Well]]],
) -> command_types.DistributeWithLiquidClassCommand:
    text = (
        "Distributing "
        + f"{volume} uL of {liquid_class.display_name} liquid class from "
        + f"{stringify_well_list(source)} to {stringify_well_list(destination)}"
    )
    return {
        "name": command_types.DISTRIBUTE_WITH_LIQUID_CLASS,
        "payload": {
            "instrument": instrument,
            "liquid_class": liquid_class,
            "volume": volume,
            "source": source,
            "destination": destination,
            "text": text,
        },
    }


def consolidate_with_liquid_class(
    instrument: InstrumentContext,
    liquid_class: LiquidClass,
    volume: float,
    source: Union[Well, Sequence[Well], Sequence[Sequence[Well]]],
    destination: Union[
        Well, Sequence[Well], Sequence[Sequence[Well]], TrashBin, WasteChute
    ],
) -> command_types.ConsolidateWithLiquidClassCommand:
    if isinstance(destination, (TrashBin, WasteChute)):
        destination_text = stringify_disposal_location(destination)
    else:
        destination_text = stringify_well_list(destination)
    text = (
        "Consolidating "
        + f"{volume} uL of {liquid_class.display_name} liquid class from "
        + f"{stringify_well_list(source)} to {destination_text}"
    )
    return {
        "name": command_types.CONSOLIDATE_WITH_LIQUID_CLASS,
        "payload": {
            "instrument": instrument,
            "liquid_class": liquid_class,
            "volume": volume,
            "source": source,
            "destination": destination,
            "text": text,
        },
    }


def seal(
    instrument: InstrumentContext,
    location: Well,
) -> command_types.SealCommand:
    location_text = stringify_location(location)
    text = f"Sealing to {location_text}"
    return {
        "name": command_types.SEAL,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def unseal(
    instrument: InstrumentContext,
    location: Well,
) -> command_types.UnsealCommand:
    location_text = stringify_location(location)
    text = f"Unsealing from {location_text}"
    return {
        "name": command_types.UNSEAL,
        "payload": {"instrument": instrument, "location": location, "text": text},
    }


def resin_tip_dispense(
    instrument: InstrumentContext,
    flow_rate: float | None,
) -> command_types.PressurizeCommand:
    if flow_rate is None:
        flow_rate = 10  # The Protocol Engine default for Resin Tip Dispense
    text = f"Pressurize pipette to dispense from resin tip at {flow_rate}uL/s."
    return {
        "name": command_types.PRESSURIZE,
        "payload": {"instrument": instrument, "text": text},
    }



import asyncio
import os
from pathlib import Path

from opentrons.config import ARCHITECTURE, SystemArchitecture
from opentrons_shared_data.errors.exceptions import CommunicationError
from opentrons_shared_data.errors.codes import ErrorCodes


class CameraException(CommunicationError):
    def __init__(self, message: str, system_error: str) -> None:
        super().__init__(
            ErrorCodes.COMMUNICATION_ERROR,
            message,
            {"internal-error-message": system_error},
        )


async def take_picture(filename: Path) -> None:
    """Take a picture and save it to filename

    :param filename: Name of file to save picture to
    :param loop: optional loop to use
    :return: None
    :raises: CameraException
    """
    try:
        os.remove(filename)
    except OSError:
        pass

    if ARCHITECTURE == SystemArchitecture.YOCTO:
        cmd = f"v4l2-ctl --device /dev/video2 --set-fmt-video=width=1280,height=720,pixelformat=MJPG --stream-mmap --stream-to={str(filename)} --stream-count=1"
    elif ARCHITECTURE == SystemArchitecture.BUILDROOT:
        cmd = f"ffmpeg -f video4linux2 -s 640x480 -i /dev/video0 -ss 0:0:1 -frames 1 {str(filename)}"
    else:  # HOST
        cmd = f'ffmpeg -f avfoundation -framerate 1  -s 640x480  -i "0" -ss 0:0:1 -frames 1 {str(filename)}'

    proc = await asyncio.create_subprocess_shell(
        cmd,
        stderr=asyncio.subprocess.PIPE,
    )

    res = await proc.stderr.read()  # type: ignore
    res = res.decode().strip()
    await proc.wait()

    if proc.returncode != 0:
        raise CameraException("Failed to communicate with camera", res)
    if not filename.exists():
        raise CameraException("Failed to save image", "")



"""Ungrip labware payload, result, and implementaiton."""

from __future__ import annotations

from opentrons.hardware_control.types import Axis
from opentrons.protocol_engine.errors.exceptions import GripperNotAttachedError
from pydantic import BaseModel
from typing import Optional, Type
from typing_extensions import Literal

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...resources import ensure_ot3_hardware

from opentrons.hardware_control import HardwareControlAPI


UnsafeUngripLabwareCommandType = Literal["unsafe/ungripLabware"]


class UnsafeUngripLabwareParams(BaseModel):
    """Payload required for an UngripLabware command."""


class UnsafeUngripLabwareResult(BaseModel):
    """Result data from the execution of an UngripLabware command."""


class UnsafeUngripLabwareImplementation(
    AbstractCommandImpl[
        UnsafeUngripLabwareParams,
        SuccessData[UnsafeUngripLabwareResult],
    ]
):
    """Ungrip labware command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api

    async def execute(
        self, params: UnsafeUngripLabwareParams
    ) -> SuccessData[UnsafeUngripLabwareResult]:
        """Ungrip Labware."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        if not ot3_hardware_api.has_gripper():
            raise GripperNotAttachedError("No gripper found to perform ungrip.")
        await ot3_hardware_api.home([Axis.G])
        return SuccessData(
            public=UnsafeUngripLabwareResult(),
        )


class UnsafeUngripLabware(
    BaseCommand[UnsafeUngripLabwareParams, UnsafeUngripLabwareResult, ErrorOccurrence]
):
    """UnsafeUngripLabware command model."""

    commandType: UnsafeUngripLabwareCommandType = "unsafe/ungripLabware"
    params: UnsafeUngripLabwareParams
    result: Optional[UnsafeUngripLabwareResult] = None

    _ImplementationCls: Type[
        UnsafeUngripLabwareImplementation
    ] = UnsafeUngripLabwareImplementation


class UnsafeUngripLabwareCreate(BaseCommandCreate[UnsafeUngripLabwareParams]):
    """UnsafeEngageAxes command request model."""

    commandType: UnsafeUngripLabwareCommandType = "unsafe/ungripLabware"
    params: UnsafeUngripLabwareParams

    _CommandCls: Type[UnsafeUngripLabware] = UnsafeUngripLabware



"""A data store of liquid classes."""

from __future__ import annotations

import dataclasses
from typing import Dict
from typing_extensions import Optional

from .. import errors
from ..actions import Action, get_state_updates
from ..types import LiquidClassRecord
from . import update_types
from ._abstract_store import HasState, HandlesActions


@dataclasses.dataclass
class LiquidClassState:
    """Our state is a bidirectional mapping between IDs <-> LiquidClassRecords."""

    # We use the bidirectional map to see if we've already assigned an ID to a liquid class when the
    # engine is asked to store a new liquid class.
    liquid_class_record_by_id: Dict[str, LiquidClassRecord]
    liquid_class_record_to_id: Dict[LiquidClassRecord, str]


class LiquidClassStore(HasState[LiquidClassState], HandlesActions):
    """Container for LiquidClassState."""

    _state: LiquidClassState

    def __init__(self) -> None:
        self._state = LiquidClassState(
            liquid_class_record_by_id={},
            liquid_class_record_to_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Update the state in response to the action."""
        for state_update in get_state_updates(action):
            if state_update.liquid_class_loaded != update_types.NO_CHANGE:
                self._handle_liquid_class_loaded_update(
                    state_update.liquid_class_loaded
                )

    def _handle_liquid_class_loaded_update(
        self, state_update: update_types.LiquidClassLoadedUpdate
    ) -> None:
        # We're just a data store. All the validation and ID generation happens in the command implementation.
        self._state.liquid_class_record_by_id[
            state_update.liquid_class_id
        ] = state_update.liquid_class_record
        self._state.liquid_class_record_to_id[
            state_update.liquid_class_record
        ] = state_update.liquid_class_id


class LiquidClassView:
    """Read-only view of the LiquidClassState."""

    _state: LiquidClassState

    def __init__(self, state: LiquidClassState) -> None:
        self._state = state

    def get(self, liquid_class_id: str) -> LiquidClassRecord:
        """Get the LiquidClassRecord with the given identifier."""
        try:
            return self._state.liquid_class_record_by_id[liquid_class_id]
        except KeyError as e:
            raise errors.LiquidClassDoesNotExistError(
                f"Liquid class ID {liquid_class_id} not found."
            ) from e

    def get_id_for_liquid_class_record(
        self, liquid_class_record: LiquidClassRecord
    ) -> Optional[str]:
        """See if the given LiquidClassRecord if already in the store, and if so, return its identifier."""
        return self._state.liquid_class_record_to_id.get(liquid_class_record)

    def get_all(self) -> Dict[str, LiquidClassRecord]:
        """Get all the LiquidClassRecords in the store."""
        return self._state.liquid_class_record_by_id.copy()



"""Union types of concrete command definitions."""

from collections.abc import Collection
from typing import Annotated, Type, Union, get_type_hints

from pydantic import Field, TypeAdapter

from opentrons.util.get_union_elements import get_union_elements

from .command import DefinedErrorData
from .pipetting_common import (
    OverpressureError,
    LiquidNotFoundError,
    TipPhysicallyAttachedError,
)
from .movement_common import StallOrCollisionError
from .flex_stacker.common import FlexStackerStallOrCollisionError

from . import absorbance_reader
from . import flex_stacker
from . import heater_shaker
from . import magnetic_module
from . import temperature_module
from . import thermocycler

from . import calibration
from . import unsafe
from . import robot

from .set_rail_lights import (
    SetRailLights,
    SetRailLightsCommandType,
    SetRailLightsCreate,
    SetRailLightsParams,
    SetRailLightsResult,
)

from .air_gap_in_place import (
    AirGapInPlace,
    AirGapInPlaceParams,
    AirGapInPlaceCreate,
    AirGapInPlaceResult,
    AirGapInPlaceCommandType,
)

from .aspirate import (
    Aspirate,
    AspirateParams,
    AspirateCreate,
    AspirateResult,
    AspirateCommandType,
)

from .aspirate_in_place import (
    AspirateInPlace,
    AspirateInPlaceParams,
    AspirateInPlaceCreate,
    AspirateInPlaceResult,
    AspirateInPlaceCommandType,
)

from .aspirate_while_tracking import (
    AspirateWhileTracking,
    AspirateWhileTrackingParams,
    AspirateWhileTrackingCreate,
    AspirateWhileTrackingResult,
    AspirateWhileTrackingCommandType,
)

from .comment import (
    Comment,
    CommentParams,
    CommentCreate,
    CommentResult,
    CommentCommandType,
)

from .custom import (
    Custom,
    CustomParams,
    CustomCreate,
    CustomResult,
    CustomCommandType,
)

from .dispense import (
    Dispense,
    DispenseParams,
    DispenseCreate,
    DispenseResult,
    DispenseCommandType,
)

from .dispense_while_tracking import (
    DispenseWhileTracking,
    DispenseWhileTrackingParams,
    DispenseWhileTrackingCreate,
    DispenseWhileTrackingResult,
    DispenseWhileTrackingCommandType,
)

from .dispense_in_place import (
    DispenseInPlace,
    DispenseInPlaceParams,
    DispenseInPlaceCreate,
    DispenseInPlaceResult,
    DispenseInPlaceCommandType,
)

from .drop_tip import (
    DropTip,
    DropTipParams,
    DropTipCreate,
    DropTipResult,
    DropTipCommandType,
)

from .drop_tip_in_place import (
    DropTipInPlace,
    DropTipInPlaceParams,
    DropTipInPlaceCreate,
    DropTipInPlaceResult,
    DropTipInPlaceCommandType,
)

from .home import (
    Home,
    HomeParams,
    HomeCreate,
    HomeResult,
    HomeCommandType,
)

from .load_labware import (
    LoadLabware,
    LoadLabwareParams,
    LoadLabwareCreate,
    LoadLabwareResult,
    LoadLabwareCommandType,
)

from .reload_labware import (
    ReloadLabware,
    ReloadLabwareParams,
    ReloadLabwareCreate,
    ReloadLabwareResult,
    ReloadLabwareCommandType,
)

from .load_liquid import (
    LoadLiquid,
    LoadLiquidParams,
    LoadLiquidCreate,
    LoadLiquidResult,
    LoadLiquidCommandType,
)

from .load_liquid_class import (
    LoadLiquidClass,
    LoadLiquidClassParams,
    LoadLiquidClassCreate,
    LoadLiquidClassResult,
    LoadLiquidClassCommandType,
)

from .load_module import (
    LoadModule,
    LoadModuleParams,
    LoadModuleCreate,
    LoadModuleResult,
    LoadModuleCommandType,
)

from .load_pipette import (
    LoadPipette,
    LoadPipetteParams,
    LoadPipetteCreate,
    LoadPipetteResult,
    LoadPipetteCommandType,
)

from .load_lid_stack import (
    LoadLidStack,
    LoadLidStackParams,
    LoadLidStackCreate,
    LoadLidStackResult,
    LoadLidStackCommandType,
)

from .load_lid import (
    LoadLid,
    LoadLidParams,
    LoadLidCreate,
    LoadLidResult,
    LoadLidCommandType,
)

from .move_labware import (
    GripperMovementError,
    MoveLabware,
    MoveLabwareParams,
    MoveLabwareCreate,
    MoveLabwareResult,
    MoveLabwareCommandType,
)

from .move_relative import (
    MoveRelative,
    MoveRelativeParams,
    MoveRelativeCreate,
    MoveRelativeResult,
    MoveRelativeCommandType,
)

from .move_to_coordinates import (
    MoveToCoordinates,
    MoveToCoordinatesParams,
    MoveToCoordinatesCreate,
    MoveToCoordinatesResult,
    MoveToCoordinatesCommandType,
)

from .move_to_well import (
    MoveToWell,
    MoveToWellParams,
    MoveToWellCreate,
    MoveToWellResult,
    MoveToWellCommandType,
)

from .move_to_addressable_area import (
    MoveToAddressableArea,
    MoveToAddressableAreaParams,
    MoveToAddressableAreaCreate,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaCommandType,
)

from .move_to_addressable_area_for_drop_tip import (
    MoveToAddressableAreaForDropTip,
    MoveToAddressableAreaForDropTipParams,
    MoveToAddressableAreaForDropTipCreate,
    MoveToAddressableAreaForDropTipResult,
    MoveToAddressableAreaForDropTipCommandType,
)

from .wait_for_resume import (
    WaitForResume,
    WaitForResumeParams,
    WaitForResumeCreate,
    WaitForResumeResult,
    WaitForResumeCommandType,
)

from .wait_for_duration import (
    WaitForDuration,
    WaitForDurationParams,
    WaitForDurationCreate,
    WaitForDurationResult,
    WaitForDurationCommandType,
)

from .pick_up_tip import (
    PickUpTip,
    PickUpTipParams,
    PickUpTipCreate,
    PickUpTipResult,
    PickUpTipCommandType,
    TipPhysicallyMissingError,
)

from .touch_tip import (
    TouchTip,
    TouchTipParams,
    TouchTipCreate,
    TouchTipResult,
    TouchTipCommandType,
)

from .save_position import (
    SavePosition,
    SavePositionParams,
    SavePositionCreate,
    SavePositionResult,
    SavePositionCommandType,
)

from .blow_out import (
    BlowOutParams,
    BlowOut,
    BlowOutCreate,
    BlowOutCommandType,
    BlowOutResult,
)

from .blow_out_in_place import (
    BlowOutInPlaceParams,
    BlowOutInPlace,
    BlowOutInPlaceCreate,
    BlowOutInPlaceCommandType,
    BlowOutInPlaceResult,
)

from .set_status_bar import (
    SetStatusBar,
    SetStatusBarParams,
    SetStatusBarCreate,
    SetStatusBarResult,
    SetStatusBarCommandType,
)

from .retract_axis import (
    RetractAxis,
    RetractAxisParams,
    RetractAxisCreate,
    RetractAxisResult,
    RetractAxisCommandType,
)

from .configure_for_volume import (
    ConfigureForVolume,
    ConfigureForVolumeParams,
    ConfigureForVolumeCreate,
    ConfigureForVolumeResult,
    ConfigureForVolumeCommandType,
)

from .prepare_to_aspirate import (
    PrepareToAspirate,
    PrepareToAspirateParams,
    PrepareToAspirateCreate,
    PrepareToAspirateResult,
    PrepareToAspirateCommandType,
)

from .configure_nozzle_layout import (
    ConfigureNozzleLayout,
    ConfigureNozzleLayoutCreate,
    ConfigureNozzleLayoutParams,
    ConfigureNozzleLayoutResult,
    ConfigureNozzleLayoutCommandType,
)

from .verify_tip_presence import (
    VerifyTipPresence,
    VerifyTipPresenceCreate,
    VerifyTipPresenceParams,
    VerifyTipPresenceResult,
    VerifyTipPresenceCommandType,
)

from .get_tip_presence import (
    GetTipPresence,
    GetTipPresenceCreate,
    GetTipPresenceParams,
    GetTipPresenceResult,
    GetTipPresenceCommandType,
)

from .get_next_tip import (
    GetNextTip,
    GetNextTipCreate,
    GetNextTipParams,
    GetNextTipResult,
    GetNextTipCommandType,
)

from .liquid_probe import (
    LiquidProbe,
    LiquidProbeParams,
    LiquidProbeCreate,
    LiquidProbeResult,
    LiquidProbeCommandType,
    TryLiquidProbe,
    TryLiquidProbeParams,
    TryLiquidProbeCreate,
    TryLiquidProbeResult,
    TryLiquidProbeCommandType,
)

from .seal_pipette_to_tip import (
    SealPipetteToTip,
    SealPipetteToTipParams,
    SealPipetteToTipCreate,
    SealPipetteToTipResult,
    SealPipetteToTipCommandType,
)

from .pressure_dispense import (
    PressureDispense,
    PressureDispenseParams,
    PressureDispenseCreate,
    PressureDispenseResult,
    PressureDispenseCommandType,
)

from .unseal_pipette_from_tip import (
    UnsealPipetteFromTip,
    UnsealPipetteFromTipParams,
    UnsealPipetteFromTipCreate,
    UnsealPipetteFromTipResult,
    UnsealPipetteFromTipCommandType,
)

Command = Annotated[
    Union[
        AirGapInPlace,
        Aspirate,
        AspirateInPlace,
        AspirateWhileTracking,
        Comment,
        Custom,
        Dispense,
        DispenseInPlace,
        DispenseWhileTracking,
        BlowOut,
        BlowOutInPlace,
        ConfigureForVolume,
        ConfigureNozzleLayout,
        DropTip,
        DropTipInPlace,
        Home,
        RetractAxis,
        LoadLabware,
        ReloadLabware,
        LoadLiquid,
        LoadLiquidClass,
        LoadModule,
        LoadPipette,
        LoadLidStack,
        LoadLid,
        MoveLabware,
        MoveRelative,
        MoveToCoordinates,
        MoveToWell,
        MoveToAddressableArea,
        MoveToAddressableAreaForDropTip,
        PrepareToAspirate,
        WaitForResume,
        WaitForDuration,
        PickUpTip,
        SavePosition,
        SetRailLights,
        TouchTip,
        SetStatusBar,
        VerifyTipPresence,
        GetTipPresence,
        GetNextTip,
        LiquidProbe,
        TryLiquidProbe,
        SealPipetteToTip,
        PressureDispense,
        UnsealPipetteFromTip,
        heater_shaker.WaitForTemperature,
        heater_shaker.SetTargetTemperature,
        heater_shaker.DeactivateHeater,
        heater_shaker.SetAndWaitForShakeSpeed,
        heater_shaker.DeactivateShaker,
        heater_shaker.OpenLabwareLatch,
        heater_shaker.CloseLabwareLatch,
        magnetic_module.Disengage,
        magnetic_module.Engage,
        temperature_module.SetTargetTemperature,
        temperature_module.WaitForTemperature,
        temperature_module.DeactivateTemperature,
        thermocycler.SetTargetBlockTemperature,
        thermocycler.WaitForBlockTemperature,
        thermocycler.SetTargetLidTemperature,
        thermocycler.WaitForLidTemperature,
        thermocycler.DeactivateBlock,
        thermocycler.DeactivateLid,
        thermocycler.OpenLid,
        thermocycler.CloseLid,
        thermocycler.RunProfile,
        thermocycler.RunExtendedProfile,
        absorbance_reader.CloseLid,
        absorbance_reader.OpenLid,
        absorbance_reader.Initialize,
        absorbance_reader.ReadAbsorbance,
        flex_stacker.Retrieve,
        flex_stacker.Store,
        flex_stacker.SetStoredLabware,
        flex_stacker.Fill,
        flex_stacker.Empty,
        flex_stacker.CloseLatch,
        flex_stacker.OpenLatch,
        flex_stacker.PrepareShuttle,
        calibration.CalibrateGripper,
        calibration.CalibratePipette,
        calibration.CalibrateModule,
        calibration.MoveToMaintenancePosition,
        unsafe.UnsafeBlowOutInPlace,
        unsafe.UnsafeDropTipInPlace,
        unsafe.UpdatePositionEstimators,
        unsafe.UnsafeEngageAxes,
        unsafe.UnsafeUngripLabware,
        unsafe.UnsafePlaceLabware,
        robot.MoveTo,
        robot.MoveAxesRelative,
        robot.MoveAxesTo,
        robot.OpenGripperJaw,
        robot.CloseGripperJaw,
    ],
    Field(discriminator="commandType"),
]

CommandParams = Union[
    AirGapInPlaceParams,
    AspirateParams,
    AspirateWhileTrackingParams,
    AspirateInPlaceParams,
    CommentParams,
    ConfigureForVolumeParams,
    ConfigureNozzleLayoutParams,
    CustomParams,
    DispenseParams,
    DispenseInPlaceParams,
    DispenseWhileTrackingParams,
    BlowOutParams,
    BlowOutInPlaceParams,
    DropTipParams,
    DropTipInPlaceParams,
    HomeParams,
    RetractAxisParams,
    LoadLabwareParams,
    LoadLidStackParams,
    LoadLidParams,
    ReloadLabwareParams,
    LoadLiquidParams,
    LoadLiquidClassParams,
    LoadModuleParams,
    LoadPipetteParams,
    MoveLabwareParams,
    MoveRelativeParams,
    MoveToCoordinatesParams,
    MoveToWellParams,
    MoveToAddressableAreaParams,
    MoveToAddressableAreaForDropTipParams,
    PrepareToAspirateParams,
    WaitForResumeParams,
    WaitForDurationParams,
    PickUpTipParams,
    SavePositionParams,
    SetRailLightsParams,
    TouchTipParams,
    SetStatusBarParams,
    VerifyTipPresenceParams,
    GetTipPresenceParams,
    GetNextTipParams,
    LiquidProbeParams,
    TryLiquidProbeParams,
    SealPipetteToTipParams,
    PressureDispenseParams,
    UnsealPipetteFromTipParams,
    heater_shaker.WaitForTemperatureParams,
    heater_shaker.SetTargetTemperatureParams,
    heater_shaker.DeactivateHeaterParams,
    heater_shaker.SetAndWaitForShakeSpeedParams,
    heater_shaker.DeactivateShakerParams,
    heater_shaker.OpenLabwareLatchParams,
    heater_shaker.CloseLabwareLatchParams,
    magnetic_module.DisengageParams,
    magnetic_module.EngageParams,
    temperature_module.SetTargetTemperatureParams,
    temperature_module.WaitForTemperatureParams,
    temperature_module.DeactivateTemperatureParams,
    thermocycler.SetTargetBlockTemperatureParams,
    thermocycler.WaitForBlockTemperatureParams,
    thermocycler.SetTargetLidTemperatureParams,
    thermocycler.WaitForLidTemperatureParams,
    thermocycler.DeactivateBlockParams,
    thermocycler.DeactivateLidParams,
    thermocycler.OpenLidParams,
    thermocycler.CloseLidParams,
    thermocycler.RunProfileParams,
    thermocycler.RunExtendedProfileParams,
    absorbance_reader.CloseLidParams,
    absorbance_reader.OpenLidParams,
    absorbance_reader.InitializeParams,
    absorbance_reader.ReadAbsorbanceParams,
    flex_stacker.RetrieveParams,
    flex_stacker.StoreParams,
    flex_stacker.SetStoredLabwareParams,
    flex_stacker.FillParams,
    flex_stacker.EmptyParams,
    flex_stacker.CloseLatchParams,
    flex_stacker.OpenLatchParams,
    flex_stacker.PrepareShuttleParams,
    calibration.CalibrateGripperParams,
    calibration.CalibratePipetteParams,
    calibration.CalibrateModuleParams,
    calibration.MoveToMaintenancePositionParams,
    unsafe.UnsafeBlowOutInPlaceParams,
    unsafe.UnsafeDropTipInPlaceParams,
    unsafe.UpdatePositionEstimatorsParams,
    unsafe.UnsafeEngageAxesParams,
    unsafe.UnsafeUngripLabwareParams,
    unsafe.UnsafePlaceLabwareParams,
    robot.MoveAxesRelativeParams,
    robot.MoveAxesToParams,
    robot.MoveToParams,
    robot.OpenGripperJawParams,
    robot.CloseGripperJawParams,
]

CommandType = Union[
    AirGapInPlaceCommandType,
    AspirateCommandType,
    AspirateWhileTrackingCommandType,
    AspirateInPlaceCommandType,
    CommentCommandType,
    ConfigureForVolumeCommandType,
    ConfigureNozzleLayoutCommandType,
    CustomCommandType,
    DispenseCommandType,
    DispenseInPlaceCommandType,
    DispenseWhileTrackingCommandType,
    BlowOutCommandType,
    BlowOutInPlaceCommandType,
    DropTipCommandType,
    DropTipInPlaceCommandType,
    HomeCommandType,
    RetractAxisCommandType,
    LoadLabwareCommandType,
    ReloadLabwareCommandType,
    LoadLiquidCommandType,
    LoadLiquidClassCommandType,
    LoadModuleCommandType,
    LoadPipetteCommandType,
    LoadLidStackCommandType,
    LoadLidCommandType,
    MoveLabwareCommandType,
    MoveRelativeCommandType,
    MoveToCoordinatesCommandType,
    MoveToWellCommandType,
    MoveToAddressableAreaCommandType,
    MoveToAddressableAreaForDropTipCommandType,
    PrepareToAspirateCommandType,
    WaitForResumeCommandType,
    WaitForDurationCommandType,
    PickUpTipCommandType,
    SavePositionCommandType,
    SetRailLightsCommandType,
    TouchTipCommandType,
    SetStatusBarCommandType,
    VerifyTipPresenceCommandType,
    GetTipPresenceCommandType,
    GetNextTipCommandType,
    LiquidProbeCommandType,
    TryLiquidProbeCommandType,
    SealPipetteToTipCommandType,
    PressureDispenseCommandType,
    UnsealPipetteFromTipCommandType,
    heater_shaker.WaitForTemperatureCommandType,
    heater_shaker.SetTargetTemperatureCommandType,
    heater_shaker.DeactivateHeaterCommandType,
    heater_shaker.SetAndWaitForShakeSpeedCommandType,
    heater_shaker.DeactivateShakerCommandType,
    heater_shaker.OpenLabwareLatchCommandType,
    heater_shaker.CloseLabwareLatchCommandType,
    magnetic_module.DisengageCommandType,
    magnetic_module.EngageCommandType,
    temperature_module.SetTargetTemperatureCommandType,
    temperature_module.WaitForTemperatureCommandType,
    temperature_module.DeactivateTemperatureCommandType,
    thermocycler.SetTargetBlockTemperatureCommandType,
    thermocycler.WaitForBlockTemperatureCommandType,
    thermocycler.SetTargetLidTemperatureCommandType,
    thermocycler.WaitForLidTemperatureCommandType,
    thermocycler.DeactivateBlockCommandType,
    thermocycler.DeactivateLidCommandType,
    thermocycler.OpenLidCommandType,
    thermocycler.CloseLidCommandType,
    thermocycler.RunProfileCommandType,
    thermocycler.RunExtendedProfileCommandType,
    absorbance_reader.CloseLidCommandType,
    absorbance_reader.OpenLidCommandType,
    absorbance_reader.InitializeCommandType,
    absorbance_reader.ReadAbsorbanceCommandType,
    flex_stacker.RetrieveCommandType,
    flex_stacker.StoreCommandType,
    flex_stacker.SetStoredLabwareCommandType,
    flex_stacker.FillCommandType,
    flex_stacker.EmptyCommandType,
    flex_stacker.CloseLatchCommandType,
    flex_stacker.OpenLatchCommandType,
    flex_stacker.PrepareShuttleCommandType,
    calibration.CalibrateGripperCommandType,
    calibration.CalibratePipetteCommandType,
    calibration.CalibrateModuleCommandType,
    calibration.MoveToMaintenancePositionCommandType,
    unsafe.UnsafeBlowOutInPlaceCommandType,
    unsafe.UnsafeDropTipInPlaceCommandType,
    unsafe.UpdatePositionEstimatorsCommandType,
    unsafe.UnsafeEngageAxesCommandType,
    unsafe.UnsafeUngripLabwareCommandType,
    unsafe.UnsafePlaceLabwareCommandType,
    robot.MoveAxesRelativeCommandType,
    robot.MoveAxesToCommandType,
    robot.MoveToCommandType,
    robot.OpenGripperJawCommandType,
    robot.CloseGripperJawCommandType,
]

CommandCreate = Annotated[
    Union[
        AirGapInPlaceCreate,
        AspirateCreate,
        AspirateWhileTrackingCreate,
        AspirateInPlaceCreate,
        CommentCreate,
        ConfigureForVolumeCreate,
        ConfigureNozzleLayoutCreate,
        CustomCreate,
        DispenseCreate,
        DispenseInPlaceCreate,
        DispenseWhileTrackingCreate,
        BlowOutCreate,
        BlowOutInPlaceCreate,
        DropTipCreate,
        DropTipInPlaceCreate,
        HomeCreate,
        RetractAxisCreate,
        LoadLabwareCreate,
        ReloadLabwareCreate,
        LoadLiquidCreate,
        LoadLiquidClassCreate,
        LoadModuleCreate,
        LoadPipetteCreate,
        LoadLidStackCreate,
        LoadLidCreate,
        MoveLabwareCreate,
        MoveRelativeCreate,
        MoveToCoordinatesCreate,
        MoveToWellCreate,
        MoveToAddressableAreaCreate,
        MoveToAddressableAreaForDropTipCreate,
        PrepareToAspirateCreate,
        WaitForResumeCreate,
        WaitForDurationCreate,
        PickUpTipCreate,
        SavePositionCreate,
        SetRailLightsCreate,
        TouchTipCreate,
        SetStatusBarCreate,
        VerifyTipPresenceCreate,
        GetTipPresenceCreate,
        GetNextTipCreate,
        LiquidProbeCreate,
        TryLiquidProbeCreate,
        SealPipetteToTipCreate,
        PressureDispenseCreate,
        UnsealPipetteFromTipCreate,
        heater_shaker.WaitForTemperatureCreate,
        heater_shaker.SetTargetTemperatureCreate,
        heater_shaker.DeactivateHeaterCreate,
        heater_shaker.SetAndWaitForShakeSpeedCreate,
        heater_shaker.DeactivateShakerCreate,
        heater_shaker.OpenLabwareLatchCreate,
        heater_shaker.CloseLabwareLatchCreate,
        magnetic_module.DisengageCreate,
        magnetic_module.EngageCreate,
        temperature_module.SetTargetTemperatureCreate,
        temperature_module.WaitForTemperatureCreate,
        temperature_module.DeactivateTemperatureCreate,
        thermocycler.SetTargetBlockTemperatureCreate,
        thermocycler.WaitForBlockTemperatureCreate,
        thermocycler.SetTargetLidTemperatureCreate,
        thermocycler.WaitForLidTemperatureCreate,
        thermocycler.DeactivateBlockCreate,
        thermocycler.DeactivateLidCreate,
        thermocycler.OpenLidCreate,
        thermocycler.CloseLidCreate,
        thermocycler.RunProfileCreate,
        thermocycler.RunExtendedProfileCreate,
        absorbance_reader.CloseLidCreate,
        absorbance_reader.OpenLidCreate,
        absorbance_reader.InitializeCreate,
        absorbance_reader.ReadAbsorbanceCreate,
        flex_stacker.RetrieveCreate,
        flex_stacker.StoreCreate,
        flex_stacker.SetStoredLabwareCreate,
        flex_stacker.FillCreate,
        flex_stacker.EmptyCreate,
        flex_stacker.CloseLatchCreate,
        flex_stacker.OpenLatchCreate,
        flex_stacker.PrepareShuttleCreate,
        calibration.CalibrateGripperCreate,
        calibration.CalibratePipetteCreate,
        calibration.CalibrateModuleCreate,
        calibration.MoveToMaintenancePositionCreate,
        unsafe.UnsafeBlowOutInPlaceCreate,
        unsafe.UnsafeDropTipInPlaceCreate,
        unsafe.UpdatePositionEstimatorsCreate,
        unsafe.UnsafeEngageAxesCreate,
        unsafe.UnsafeUngripLabwareCreate,
        unsafe.UnsafePlaceLabwareCreate,
        robot.MoveAxesRelativeCreate,
        robot.MoveAxesToCreate,
        robot.MoveToCreate,
        robot.OpenGripperJawCreate,
        robot.CloseGripperJawCreate,
    ],
    Field(discriminator="commandType"),
]

# Each time a TypeAdapter is instantiated, it will construct a new validator and
# serializer. To improve performance, TypeAdapters are instantiated once.
# See https://docs.pydantic.dev/latest/concepts/performance/#typeadapter-instantiated-once
CommandCreateAdapter: TypeAdapter[CommandCreate] = TypeAdapter(CommandCreate)

CommandAdapter: TypeAdapter[Command] = TypeAdapter(Command)

CommandResult = Union[
    AirGapInPlaceResult,
    AspirateResult,
    AspirateWhileTrackingResult,
    AspirateInPlaceResult,
    CommentResult,
    ConfigureForVolumeResult,
    ConfigureNozzleLayoutResult,
    CustomResult,
    DispenseResult,
    DispenseInPlaceResult,
    DispenseWhileTrackingResult,
    BlowOutResult,
    BlowOutInPlaceResult,
    DropTipResult,
    DropTipInPlaceResult,
    HomeResult,
    RetractAxisResult,
    LoadLabwareResult,
    ReloadLabwareResult,
    LoadLiquidResult,
    LoadLiquidClassResult,
    LoadModuleResult,
    LoadPipetteResult,
    LoadLidStackResult,
    LoadLidResult,
    MoveLabwareResult,
    MoveRelativeResult,
    MoveToCoordinatesResult,
    MoveToWellResult,
    MoveToAddressableAreaResult,
    MoveToAddressableAreaForDropTipResult,
    PrepareToAspirateResult,
    WaitForResumeResult,
    WaitForDurationResult,
    PickUpTipResult,
    SavePositionResult,
    SetRailLightsResult,
    TouchTipResult,
    SetStatusBarResult,
    VerifyTipPresenceResult,
    GetTipPresenceResult,
    GetNextTipResult,
    LiquidProbeResult,
    TryLiquidProbeResult,
    SealPipetteToTipResult,
    PressureDispenseResult,
    UnsealPipetteFromTipResult,
    heater_shaker.WaitForTemperatureResult,
    heater_shaker.SetTargetTemperatureResult,
    heater_shaker.DeactivateHeaterResult,
    heater_shaker.SetAndWaitForShakeSpeedResult,
    heater_shaker.DeactivateShakerResult,
    heater_shaker.OpenLabwareLatchResult,
    heater_shaker.CloseLabwareLatchResult,
    magnetic_module.DisengageResult,
    magnetic_module.EngageResult,
    temperature_module.SetTargetTemperatureResult,
    temperature_module.WaitForTemperatureResult,
    temperature_module.DeactivateTemperatureResult,
    thermocycler.SetTargetBlockTemperatureResult,
    thermocycler.WaitForBlockTemperatureResult,
    thermocycler.SetTargetLidTemperatureResult,
    thermocycler.WaitForLidTemperatureResult,
    thermocycler.DeactivateBlockResult,
    thermocycler.DeactivateLidResult,
    thermocycler.OpenLidResult,
    thermocycler.CloseLidResult,
    thermocycler.RunProfileResult,
    thermocycler.RunExtendedProfileResult,
    absorbance_reader.CloseLidResult,
    absorbance_reader.OpenLidResult,
    absorbance_reader.InitializeResult,
    absorbance_reader.ReadAbsorbanceResult,
    flex_stacker.RetrieveResult,
    flex_stacker.StoreResult,
    flex_stacker.SetStoredLabwareResult,
    flex_stacker.FillResult,
    flex_stacker.EmptyResult,
    flex_stacker.CloseLatchResult,
    flex_stacker.OpenLatchResult,
    flex_stacker.PrepareShuttleResult,
    calibration.CalibrateGripperResult,
    calibration.CalibratePipetteResult,
    calibration.CalibrateModuleResult,
    calibration.MoveToMaintenancePositionResult,
    unsafe.UnsafeBlowOutInPlaceResult,
    unsafe.UnsafeDropTipInPlaceResult,
    unsafe.UpdatePositionEstimatorsResult,
    unsafe.UnsafeEngageAxesResult,
    unsafe.UnsafeUngripLabwareResult,
    unsafe.UnsafePlaceLabwareResult,
    robot.MoveAxesRelativeResult,
    robot.MoveAxesToResult,
    robot.MoveToResult,
    robot.OpenGripperJawResult,
    robot.CloseGripperJawResult,
]


# All `DefinedErrorData`s that implementations will actually return in practice.
CommandDefinedErrorData = Union[
    DefinedErrorData[TipPhysicallyMissingError],
    DefinedErrorData[TipPhysicallyAttachedError],
    DefinedErrorData[OverpressureError],
    DefinedErrorData[LiquidNotFoundError],
    DefinedErrorData[GripperMovementError],
    DefinedErrorData[StallOrCollisionError],
    DefinedErrorData[FlexStackerStallOrCollisionError],
]


def _map_create_types_by_params_type(
    create_types: Collection[Type[CommandCreate]],
) -> dict[Type[CommandParams], Type[CommandCreate]]:
    def get_params_type(create_type: Type[CommandCreate]) -> Type[CommandParams]:
        return get_type_hints(create_type)["params"]  # type: ignore[no-any-return]

    result = {get_params_type(create_type): create_type for create_type in create_types}

    # This isn't an inherent requirement of opentrons.protocol_engine,
    # but this mapping is only useful to higher-level code if this holds true.
    assert len(result) == len(
        create_types
    ), "Param models should map to create models 1:1."

    return result


CREATE_TYPES_BY_PARAMS_TYPE = _map_create_types_by_params_type(
    get_union_elements(CommandCreate)
)
"""A "reverse" mapping from each CommandParams type to its parent CommandCreate type."""



from __future__ import annotations
import logging
import numpy as np
from numpy import insert, dot
from numpy.linalg import inv
from numpy.typing import NDArray
from typing import List, Tuple, Union

from opentrons.calibration_storage.types import AttitudeMatrix

mod_log = logging.getLogger(__name__)

# (TODO(lc, 8/11/2020): temporary type until
# old calibration data is removed.
AxisPosition = Union[Tuple[float, float, float], Tuple[float, float]]

SolvePoints = Tuple[
    Tuple[float, float, float], Tuple[float, float, float], Tuple[float, float, float]
]

DoubleArray = NDArray[np.double]
DoubleMatrix = NDArray[np.double]


def identity_deck_transform() -> DoubleArray:
    """The default deck transform"""
    return np.identity(3)


def solve_attitude(expected: SolvePoints, actual: SolvePoints) -> AttitudeMatrix:
    ex: DoubleMatrix = np.array([list(point) for point in expected]).transpose()
    ac: DoubleMatrix = np.array([list(point) for point in actual]).transpose()
    t = np.dot(ac, inv(ex))

    mask_transform: NDArray[np.bool_] = np.array(
        [[True, True, False], [True, True, False], [False, False, False]]
    )
    masked_array = np.ma.masked_array(t, ~mask_transform)  # type: ignore[var-annotated, no-untyped-call]

    no_z_component = np.zeros((3, 3))
    np.put(no_z_component, [8, 8], 1)

    transform = masked_array.filled(0) + no_z_component  # type: ignore[no-untyped-call]
    return transform.round(4).tolist()  # type: ignore[no-any-return]


def solve(
    expected: List[Tuple[float, float]],
    actual: List[Tuple[float, float]],
) -> DoubleArray:
    """
    Takes two lists of 3 x-y points each, and calculates the matrix
    representing the transformation from one space to the other.

    The 3x3 matrix returned by this method represents the 2-D transformation
    matrix from the actual point to the expected point.

    Example:
        If the expected points are:
            [ (1, 1),
              (2, 2),
              (1, 2) ]
        And the actual measured points are:
            [ (1.1, 1.1),
              (2.1, 2.1),
              (1.1, 2.1) ]
        (in other words, a shift of exaxtly +0.1 in both x and y)

        Then the resulting transformation matrix T should be:
            [ 1  0  -0.1 ]
            [ 0  1  -0.1 ]
            [ 0  0   1   ]

        Then, if we take a 3x3 matrix B representing one of the measured points
        on the deck:
            [ 1  0  1.1 ]
            [ 0  1  2.1 ]
            [ 0  0  1   ]

        The B*T will yield the "actual" point:
            [ 1  0  1 ]
            [ 0  1  2 ]
            [ 0  0  1 ]

        The return value of this function is the transformation matrix T
    """
    # Note: input list shape validation is handled by the type checker

    # Turn expected and actual matricies into numpy ndarrays with the last row
    # of [1 1 1] appended, and then take the dot product of the resulting
    # actual matrix with the inverse of the resulting expected matrix.

    # Shape of `expected` and `actual`:
    # [ (x1, y1),
    #   (x2, y2),
    #   (x3, y3) ]
    ex: DoubleMatrix = np.array([list(point) + [1] for point in expected]).transpose()

    ac: DoubleMatrix = np.array([list(point) + [1] for point in actual]).transpose()
    # Shape of `ex` and `ac`:
    # [ x1 x2 x3 ]
    # [ y1 y2 y3 ]
    # [  1  1  1 ]

    transform = np.dot(ac, inv(ex))
    # `dot` in numpy is a misnomer. When both arguments are square, N-
    # dimensional arrays, the return type is the result of performing matrix
    # multiplication, rather than the dot-product (so the return here will be
    # a 4x4 matrix)
    return transform  # type: ignore[no-any-return]


def add_z(xy: DoubleArray, z: float) -> DoubleArray:
    """
    Turn a 2-D transform matrix into a 3-D transform matrix (scale/shift only,
    no rotation).

    :param xy: A two-dimensional transform matrix (a 3x3 numpy ndarray) in the
        following form:
            [ 1  0  x ]
            [ 0  1  y ]
            [ 0  0  1 ]
    :param z: a float for the z component
    :return: a three-dimensional transformation matrix (a 4x4 numpy ndarray)
        with x, y, and z from the function parameters, in the following form:
            [ 1  0  0  x ]
            [ 0  1  0  y ]
            [ 0  0  1  z ]
            [ 0  0  0  1 ]
    """
    # First, insert a column of zeros as into the input matrix
    interm = insert(xy, 2, [0, 0, 0], axis=1)
    # Result:
    # [ 1  0  0  x ]
    # [ 0  1  0  y ]
    # [ 0  0  0  1 ]

    # Then, insert the z row to create a properly formed 3-D transform matrix:
    xyz: DoubleMatrix = insert(interm, 2, [0, 0, 1, z], axis=0)
    # Result:
    # [ 1  0  0  x ]
    # [ 0  1  0  y ]
    # [ 0  0  1  z ]
    # [ 0  0  0  1 ]

    return xyz.round(11)


def add_matrices(
    t1: Tuple[float, float, float], t2: Tuple[float, float, float]
) -> Tuple[float, float, float]:
    """
    Simple method to convert tuples to numpy arrays and add them.
    """
    return tuple(np.asarray(t1) + np.asarray(t2))


def apply_transform(
    t: Union[List[List[float]], DoubleArray],
    pos: AxisPosition,
) -> Tuple[float, float, float]:
    """
    Change of base using a transform matrix. Primarily used to render a point
    in space in a way that is more readable for the user.

    :param t: A transformation matrix from one 3D space [A] to another [B]
    :param pos: XYZ point in space A
    :return: corresponding XYZ point in space B
    """
    return tuple(dot(t, list(pos))[:3])


def apply_reverse(
    t: Union[List[List[float]], DoubleArray],
    pos: AxisPosition,
) -> Tuple[float, float, float]:
    """Like apply_transform but inverts the transform first"""
    return apply_transform(inv(t), pos)



"""Public protocol engine value types and models."""

from __future__ import annotations


from opentrons_shared_data.pipette.types import LabwareUri
from opentrons.hardware_control.modules import ModuleType


from .run_time_parameters import (
    NumberParameter,
    BooleanParameter,
    EnumParameter,
    CSVParameter,
    RunTimeParameter,
    PrimitiveRunTimeParamValuesType,
    CSVRunTimeParamFilesType,
    CSVRuntimeParamPaths,
    FileInfo,
    EnumChoice,
)

from .command_annotations import (
    SecondOrderCommandAnnotation,
    CustomCommandAnnotation,
    CommandAnnotation,
)
from .partial_tip_configuration import (
    AllNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
    RowNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    QuadrantNozzleLayoutConfiguration,
    NozzleLayoutConfigurationType,
    PRIMARY_NOZZLE_LITERAL,
)
from .automatic_tip_selection import NextTipInfo, NoTipReason, NoTipAvailable
from .instrument_sensors import InstrumentSensorId, TipPresenceStatus
from .deck_configuration import (
    AddressableOffsetVector,
    PotentialCutoutFixture,
    AreaType,
    AddressableArea,
    DeckConfigurationType,
    DeckType,
)
from .liquid_class import LiquidClassRecord, LiquidClassRecordWithId
from .module import (
    ModuleModel,
    TemperatureModuleModel,
    MagneticModuleModel,
    ThermocyclerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
    AbsorbanceReaderModel,
    FlexStackerModuleModel,
    ModuleDimensions,
    ModuleCalibrationPoint,
    ModuleDefinition,
    LoadedModule,
    SpeedRange,
    TemperatureRange,
    HeaterShakerLatchStatus,
    HeaterShakerMovementRestrictors,
    ABSMeasureMode,
    ModuleOffsetVector,
    ModuleOffsetData,
    StackerFillEmptyStrategy,
)
from .location import (
    DeckSlotLocation,
    StagingSlotLocation,
    AddressableAreaLocation,
    ModuleLocation,
    OnLabwareLocation,
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
    LabwareLocation,
    OnDeckLabwareLocation,
    NonStackedLocation,
    DeckPoint,
    InStackerHopperLocation,
    OnLabwareLocationSequenceComponent,
    OnModuleLocationSequenceComponent,
    OnAddressableAreaLocationSequenceComponent,
    NotOnDeckLocationSequenceComponent,
    OnCutoutFixtureLocationSequenceComponent,
    LabwareLocationSequence,
    LoadableLabwareLocation,
    labware_location_is_system,
    labware_location_is_off_deck,
)
from .labware import (
    OverlapOffset,
    LabwareOffset,
    LabwareOffsetCreate,
    LegacyLabwareOffsetCreate,
    LabwareOffsetCreateInternal,
    LoadedLabware,
)
from .liquid import HexColor, EmptyLiquidId, LiquidId, Liquid, FluidKind, AspiratedFluid
from .labware_offset_location import (
    LegacyLabwareOffsetLocation,
    LabwareOffsetLocationSequence,
    OnLabwareOffsetLocationSequenceComponent,
    OnModuleOffsetLocationSequenceComponent,
    OnAddressableAreaOffsetLocationSequenceComponent,
    LabwareOffsetLocationSequenceComponents,
)
from .labware_offset_vector import LabwareOffsetVector
from .well_position import (
    WellOrigin,
    PickUpTipWellOrigin,
    DropTipWellOrigin,
    WellOffset,
    WellLocation,
    LiquidHandlingWellLocation,
    PickUpTipWellLocation,
    DropTipWellLocation,
    WellLocationType,
    WellLocationFunction,
)
from .instrument import (
    LoadedPipette,
    CurrentAddressableArea,
    CurrentWell,
    CurrentPipetteLocation,
    InstrumentOffsetVector,
)
from .execution import EngineStatus, PostRunHardwareState
from .liquid_level_detection import (
    LoadedVolumeInfo,
    ProbedHeightInfo,
    ProbedVolumeInfo,
    WellInfoSummary,
    WellLiquidInfo,
    LiquidTrackingType,
    SimulatedProbeResult,
)
from .liquid_handling import FlowRates
from .labware_movement import LabwareMovementStrategy, LabwareMovementOffsetData
from .tip import TipGeometry
from .hardware_passthrough import MovementAxis, MotorAxis
from .util import Vec3f, Dimensions

__all__ = [
    # Runtime parameters
    "NumberParameter",
    "BooleanParameter",
    "EnumParameter",
    "EnumChoice",
    "CSVParameter",
    "PrimitiveRunTimeParamValuesType",
    "CSVRunTimeParamFilesType",
    "CSVRuntimeParamPaths",
    "FileInfo",
    "RunTimeParameter",
    # Command annotations
    "SecondOrderCommandAnnotation",
    "CustomCommandAnnotation",
    "CommandAnnotation",
    # Partial tip handling
    "AllNozzleLayoutConfiguration",
    "SingleNozzleLayoutConfiguration",
    "RowNozzleLayoutConfiguration",
    "ColumnNozzleLayoutConfiguration",
    "QuadrantNozzleLayoutConfiguration",
    "NozzleLayoutConfigurationType",
    "PRIMARY_NOZZLE_LITERAL",
    # Automatic tip selection
    "NextTipInfo",
    "NoTipReason",
    "NoTipAvailable",
    # Instrument sensors
    "InstrumentSensorId",
    "TipPresenceStatus",
    # Deck configuration
    "AddressableOffsetVector",
    "PotentialCutoutFixture",
    "AreaType",
    "AddressableArea",
    "DeckConfigurationType",
    "DeckType",
    # Liquid classes
    "LiquidClassRecord",
    "LiquidClassRecordWithId",
    # Modules
    "ModuleModel",
    "ModuleType",
    "TemperatureModuleModel",
    "MagneticModuleModel",
    "ThermocyclerModuleModel",
    "HeaterShakerModuleModel",
    "MagneticBlockModel",
    "AbsorbanceReaderModel",
    "FlexStackerModuleModel",
    "ModuleDimensions",
    "ModuleCalibrationPoint",
    "ModuleDefinition",
    "LoadedModule",
    "SpeedRange",
    "TemperatureRange",
    "HeaterShakerLatchStatus",
    "HeaterShakerMovementRestrictors",
    "ABSMeasureMode",
    "ModuleOffsetVector",
    "ModuleOffsetData",
    "StackerFillEmptyStrategy",
    # Locations of things on deck
    "DeckSlotLocation",
    "StagingSlotLocation",
    "AddressableAreaLocation",
    "ModuleLocation",
    "OnLabwareLocation",
    "OFF_DECK_LOCATION",
    "SYSTEM_LOCATION",
    "LabwareLocation",
    "OnDeckLabwareLocation",
    "NonStackedLocation",
    "DeckPoint",
    "OffDeckLocationType",
    "SystemLocationType",
    "InStackerHopperLocation",
    "OnLabwareLocationSequenceComponent",
    "OnModuleLocationSequenceComponent",
    "OnAddressableAreaLocationSequenceComponent",
    "NotOnDeckLocationSequenceComponent",
    "OnCutoutFixtureLocationSequenceComponent",
    "LabwareLocationSequence",
    "LoadableLabwareLocation",
    "labware_location_is_off_deck",
    "labware_location_is_system",
    # Labware offset location
    "LegacyLabwareOffsetLocation",
    "LabwareOffsetLocationSequence",
    "LabwareOffsetLocationSequenceComponents",
    "OnLabwareOffsetLocationSequenceComponent",
    "OnModuleOffsetLocationSequenceComponent",
    "OnAddressableAreaOffsetLocationSequenceComponent",
    # Labware offset vector
    "LabwareOffsetVector",
    # Labware
    "OverlapOffset",
    "LabwareOffset",
    "LabwareOffsetCreate",
    "LegacyLabwareOffsetCreate",
    "LabwareOffsetCreateInternal",
    "LoadedLabware",
    "LabwareOffsetVector",
    # Liquids
    "HexColor",
    "EmptyLiquidId",
    "LiquidId",
    "Liquid",
    "FluidKind",
    "AspiratedFluid",
    # Well locations
    "WellOrigin",
    "PickUpTipWellOrigin",
    "DropTipWellOrigin",
    "WellOffset",
    "WellLocation",
    "LiquidHandlingWellLocation",
    "PickUpTipWellLocation",
    "DropTipWellLocation",
    "WellLocationType",
    "WellLocationFunction",
    # Execution
    "EngineStatus",
    "PostRunHardwareState",
    # Instruments
    "LoadedPipette",
    "CurrentAddressableArea",
    "CurrentWell",
    "CurrentPipetteLocation",
    "InstrumentOffsetVector",
    # Liquid level detection types
    "LoadedVolumeInfo",
    "ProbedHeightInfo",
    "ProbedVolumeInfo",
    "WellInfoSummary",
    "WellLiquidInfo",
    "LiquidTrackingType",
    "SimulatedProbeResult",
    # Liquid handling
    "FlowRates",
    # Labware movement
    "LabwareMovementStrategy",
    "LabwareMovementOffsetData",
    # Tips
    "TipGeometry",
    # Hardware passthrough
    "MovementAxis",
    "MotorAxis",
    # Utility types
    "Vec3f",
    "Dimensions",
    # Convenience re-export
    "LabwareUri",
]



"""Types for an APIv2 protocol to report its activity.

This module is not part of the public Python Protocol API.
It's only for internal Opentrons use.
"""


from dataclasses import dataclass
from typing import Optional, Union
from opentrons_shared_data.labware.types import LabwareDefinition2

from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.modules.types import ModuleModel
from opentrons.types import Mount, DeckSlotName


@dataclass(frozen=True)
class LabwareLoadInfo:
    """Information about a successful labware load.

    This is a separate class from the main user-facing `Labware` class
    because this is easier to construct in unit tests.
    """

    labware_definition: LabwareDefinition2

    # todo(mm, 2021-10-11): Namespace, load name, and version can be derived from the
    # definition. Should they be removed from here?
    labware_namespace: str
    labware_load_name: str
    labware_version: int

    # If on_module is True, deck_slot is the slot occupied by the module that the
    # labware is on.
    deck_slot: DeckSlotName
    on_module: bool

    # The ID of the labware offset resource that applied to this labware load,
    # if there was one.
    offset_id: Optional[str]

    # user-specified label if present
    labware_display_name: Optional[str]


@dataclass(frozen=True)
class InstrumentLoadInfo:
    """Like `LabwareLoadInfo`, but for instruments (pipettes)."""

    instrument_load_name: str
    mount: Mount
    pipette_dict: PipetteDict


@dataclass(frozen=True)
class ModuleLoadInfo:
    """Like `LabwareLoadInfo`, but for hardware modules."""

    requested_model: ModuleModel
    loaded_model: ModuleModel

    deck_slot: DeckSlotName
    configuration: Optional[str]
    module_serial: str


LoadInfo = Union[LabwareLoadInfo, InstrumentLoadInfo, ModuleLoadInfo]



"""A Protocol-Engine-friendly wrapper for opentrons.motion_planning.deck_conflict."""
from __future__ import annotations
import logging
from typing import (
    Optional,
    Tuple,
    Union,
    List,
)

from opentrons_shared_data.errors.exceptions import MotionPlanningFailureError
from opentrons.protocol_engine.errors import LocationIsStagingSlotError
from opentrons_shared_data.module import FLEX_TC_LID_COLLISION_ZONE

from opentrons.hardware_control import CriticalPoint
from opentrons.motion_planning import adjacent_slots_getters

from opentrons.protocol_engine import (
    StateView,
    DeckSlotLocation,
    OnLabwareLocation,
    DropTipWellLocation,
)
from opentrons.protocol_engine.types import StagingSlotLocation, WellLocationType
from opentrons.types import DeckSlotName, StagingSlotName, Point
from . import point_calculations


class PartialTipMovementNotAllowedError(MotionPlanningFailureError):
    """Error raised when trying to perform a partial tip movement to an illegal location."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message=message,
        )


class UnsuitableTiprackForPipetteMotion(MotionPlanningFailureError):
    """Error raised when trying to perform a pipette movement to a tip rack, based on adapter status."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message=message,
        )


_log = logging.getLogger(__name__)

_FLEX_TC_LID_BACK_LEFT_PT = Point(
    x=FLEX_TC_LID_COLLISION_ZONE["back_left"]["x"],
    y=FLEX_TC_LID_COLLISION_ZONE["back_left"]["y"],
    z=FLEX_TC_LID_COLLISION_ZONE["back_left"]["z"],
)

_FLEX_TC_LID_FRONT_RIGHT_PT = Point(
    x=FLEX_TC_LID_COLLISION_ZONE["front_right"]["x"],
    y=FLEX_TC_LID_COLLISION_ZONE["front_right"]["y"],
    z=FLEX_TC_LID_COLLISION_ZONE["front_right"]["z"],
)


def check_safe_for_pipette_movement(  # noqa: C901
    engine_state: StateView,
    pipette_id: str,
    labware_id: str,
    well_name: str,
    well_location: WellLocationType,
) -> None:
    """Check if the labware is safe to move to with a pipette in partial tip configuration.

    Args:
        engine_state: engine state view
        pipette_id: ID of the pipette to be moved
        labware_id: ID of the labware we are moving to
        well_name: Name of the well to move to
        well_location: exact location within the well to move to
    """
    # TODO (spp, 2023-02-06): remove this check after thorough testing.
    #  This function is capable of checking for movement conflict regardless of
    #  nozzle configuration.
    if not engine_state.pipettes.get_is_partially_configured(pipette_id):
        return

    if isinstance(well_location, DropTipWellLocation):
        # convert to WellLocation
        well_location = engine_state.geometry.get_checked_tip_drop_location(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_location=well_location,
            partially_configured=True,
        )
    well_location_point = engine_state.geometry.get_well_position(
        labware_id=labware_id,
        well_name=well_name,
        well_location=well_location,
        pipette_id=pipette_id,
    )
    primary_nozzle = engine_state.pipettes.get_primary_nozzle(pipette_id)

    destination_cp = _get_critical_point_to_use(engine_state, labware_id)
    pipette_bounds_at_well_location = (
        engine_state.pipettes.get_pipette_bounds_at_specified_move_to_position(
            pipette_id=pipette_id,
            destination_position=well_location_point,
            critical_point=destination_cp,
        )
    )
    if not _is_within_pipette_extents(
        engine_state=engine_state,
        pipette_id=pipette_id,
        pipette_bounding_box_at_loc=pipette_bounds_at_well_location,
    ):
        raise PartialTipMovementNotAllowedError(
            f"Requested motion with the {primary_nozzle} nozzle partial configuration"
            f" is outside of robot bounds for the pipette."
        )
    ancestor = engine_state.geometry.get_ancestor_slot_name(labware_id)
    if isinstance(ancestor, StagingSlotName):
        raise LocationIsStagingSlotError(
            "Cannot perform pipette actions on labware in Staging Area Slot."
        )
    labware_slot = ancestor

    surrounding_slots = adjacent_slots_getters.get_surrounding_slots(
        slot=labware_slot.as_int(), robot_type=engine_state.config.robot_type
    )

    if _will_collide_with_thermocycler_lid(
        engine_state=engine_state,
        pipette_bounds=pipette_bounds_at_well_location,
        surrounding_regular_slots=surrounding_slots.regular_slots,
    ):
        raise PartialTipMovementNotAllowedError(
            f"Moving to {engine_state.labware.get_display_name(labware_id)} in slot"
            f" {labware_slot} with {primary_nozzle} nozzle partial configuration"
            f" will result in collision with thermocycler lid in deck slot A1."
        )

    for regular_slot in surrounding_slots.regular_slots:
        if _slot_has_potential_colliding_object(
            engine_state=engine_state,
            pipette_bounds=pipette_bounds_at_well_location,
            surrounding_slot=regular_slot,
        ):
            raise PartialTipMovementNotAllowedError(
                f"Moving to {engine_state.labware.get_display_name(labware_id)} in slot"
                f" {labware_slot} with {primary_nozzle} nozzle partial configuration"
                f" will result in collision with items in deck slot {regular_slot}."
            )
    for staging_slot in surrounding_slots.staging_slots:
        if _slot_has_potential_colliding_object(
            engine_state=engine_state,
            pipette_bounds=pipette_bounds_at_well_location,
            surrounding_slot=staging_slot,
        ):
            raise PartialTipMovementNotAllowedError(
                f"Moving to {engine_state.labware.get_display_name(labware_id)} in slot"
                f" {labware_slot} with {primary_nozzle} nozzle partial configuration"
                f" will result in collision with items in staging slot {staging_slot}."
            )


def _get_critical_point_to_use(
    engine_state: StateView, labware_id: str
) -> Optional[CriticalPoint]:
    """Return the critical point to use when accessing the given labware."""
    # TODO (spp, 2024-09-17): looks like Y_CENTER of column is the same as its XY_CENTER.
    #   I'm using this if-else ladder to be consistent with what we do in
    #   `MotionPlanning.get_movement_waypoints_to_well()`.
    #   We should probably use only XY_CENTER in both places.
    if engine_state.labware.get_should_center_column_on_target_well(labware_id):
        return CriticalPoint.Y_CENTER
    elif engine_state.labware.get_should_center_pipette_on_target_well(labware_id):
        return CriticalPoint.XY_CENTER
    return None


def _slot_has_potential_colliding_object(
    engine_state: StateView,
    pipette_bounds: Tuple[Point, Point, Point, Point],
    surrounding_slot: Union[DeckSlotName, StagingSlotName],
) -> bool:
    """Return the slot, if any, that has an item that the pipette might collide into."""
    # Check if slot overlaps with pipette position
    slot_pos = engine_state.addressable_areas.get_addressable_area_position(
        addressable_area_name=surrounding_slot.id,
        do_compatibility_check=False,
    )
    slot_bounds = engine_state.addressable_areas.get_addressable_area_bounding_box(
        addressable_area_name=surrounding_slot.id,
        do_compatibility_check=False,
    )
    slot_back_left_coords = Point(slot_pos.x, slot_pos.y + slot_bounds.y, slot_pos.z)
    slot_front_right_coords = Point(slot_pos.x + slot_bounds.x, slot_pos.y, slot_pos.z)

    # If slot overlaps with pipette bounds
    if point_calculations.are_overlapping_rectangles(
        rectangle1=(pipette_bounds[0], pipette_bounds[1]),
        rectangle2=(slot_back_left_coords, slot_front_right_coords),
    ):
        # Check z-height of items in overlapping slot
        if isinstance(surrounding_slot, DeckSlotName):
            slot_highest_z = engine_state.geometry.get_highest_z_in_slot(
                DeckSlotLocation(slotName=surrounding_slot)
            )
        else:
            slot_highest_z = engine_state.geometry.get_highest_z_in_slot(
                StagingSlotLocation(slotName=surrounding_slot)
            )
        return slot_highest_z >= pipette_bounds[0].z
    return False


def _will_collide_with_thermocycler_lid(
    engine_state: StateView,
    pipette_bounds: Tuple[Point, Point, Point, Point],
    surrounding_regular_slots: List[DeckSlotName],
) -> bool:
    """Return whether the pipette might collide with thermocycler's lid/clips on a Flex.

    If any of the pipette's bounding vertices lie inside the no-go zone of the thermocycler-
    which is the area that's to the left, back and below the thermocycler's lid's
    protruding clips, then we will mark the movement for possible collision.

    This could cause false raises for the case where an 8-channel is accessing the
    thermocycler labware in a location such that the pipette is in the area between
    the clips but not touching either clips. But that's a tradeoff we'll need to make
    between a complicated check involving accurate positions of all entities involved
    and a crude check that disallows all partial tip movements around the thermocycler.
    """
    # TODO (spp, 2024-02-27): Improvements:
    #  - make the check dynamic according to lid state:
    #     - if lid is open, check if pipette is in no-go zone
    #     - if lid is closed, use the closed lid height to check for conflict
    if (
        DeckSlotName.SLOT_A1 in surrounding_regular_slots
        and engine_state.modules.is_flex_deck_with_thermocycler()
    ):
        return (
            point_calculations.are_overlapping_rectangles(
                rectangle1=(_FLEX_TC_LID_BACK_LEFT_PT, _FLEX_TC_LID_FRONT_RIGHT_PT),
                rectangle2=(pipette_bounds[0], pipette_bounds[1]),
            )
            and pipette_bounds[0].z <= _FLEX_TC_LID_BACK_LEFT_PT.z
        )

    return False


def check_safe_for_tip_pickup_and_return(
    engine_state: StateView,
    pipette_id: str,
    labware_id: str,
) -> None:
    """Check if the presence or absence of a tiprack adapter might cause any pipette movement issues.

    A 96 channel pipette will pick up tips using cam action when it's configured
    to use ALL nozzles. For this, the tiprack needs to be on the Flex 96 channel tiprack adapter
    or similar or the tips will not be picked up.

    On the other hand, if the pipette is configured with partial nozzle configuration,
    it uses the usual pipette presses to pick the tips up, in which case, having the tiprack
    on the Flex 96 channel tiprack adapter (or similar) will cause the pipette to
    crash against the adapter posts.

    In order to check if the 96-channel can move and pickup/drop tips safely, this method
    checks for the height attribute of the tiprack adapter rather than checking for the
    specific official adapter since users might create custom labware &/or definitions
    compatible with the official adapter.
    """
    if not engine_state.pipettes.get_channels(pipette_id) == 96:
        # Adapters only matter to 96 ch.
        return

    is_partial_config = engine_state.pipettes.get_is_partially_configured(pipette_id)
    tiprack_name = engine_state.labware.get_display_name(labware_id)
    tiprack_parent = engine_state.labware.get_location(labware_id)
    if isinstance(tiprack_parent, OnLabwareLocation):  # tiprack is on an adapter
        is_96_ch_tiprack_adapter = engine_state.labware.get_has_quirk(
            labware_id=tiprack_parent.labwareId, quirk="tiprackAdapterFor96Channel"
        )
        tiprack_height = engine_state.labware.get_dimensions(labware_id=labware_id).z
        adapter_height = engine_state.labware.get_dimensions(
            labware_id=tiprack_parent.labwareId
        ).z
        if is_partial_config and tiprack_height < adapter_height:
            raise PartialTipMovementNotAllowedError(
                f"{tiprack_name} cannot be on an adapter taller than the tip rack"
                f" when picking up fewer than 96 tips."
            )
        elif not is_partial_config and not is_96_ch_tiprack_adapter:
            raise UnsuitableTiprackForPipetteMotion(
                f"{tiprack_name} must be on an Opentrons Flex 96 Tip Rack Adapter"
                f" in order to pick up or return all 96 tips simultaneously."
            )

    elif (
        not is_partial_config
    ):  # tiprack is not on adapter and pipette is in full config
        raise UnsuitableTiprackForPipetteMotion(
            f"{tiprack_name} must be on an Opentrons Flex 96 Tip Rack Adapter"
            f" in order to pick up or return all 96 tips simultaneously."
        )


def _is_within_pipette_extents(
    engine_state: StateView,
    pipette_id: str,
    pipette_bounding_box_at_loc: Tuple[Point, Point, Point, Point],
) -> bool:
    """Whether a given point is within the extents of a configured pipette on the specified robot."""
    channels = engine_state.pipettes.get_channels(pipette_id)
    robot_extents = engine_state.geometry.absolute_deck_extents
    (
        pip_back_left_bound,
        pip_front_right_bound,
        pip_back_right_bound,
        pip_front_left_bound,
    ) = pipette_bounding_box_at_loc

    # Given the padding values accounted for against the deck extents,
    # a pipette is within extents when all of the following are true:

    # Each corner slot full pickup case:
    # A1: Front right nozzle is within the rear and left-side padding limits
    # D1: Back right nozzle is within the front and left-side padding limits
    # A3 Front left nozzle is within the rear and right-side padding limits
    # D3: Back left nozzle is within the front and right-side padding limits
    # Thermocycler Column A2: Front right nozzle is within padding limits

    if channels == 96:
        return (
            pip_front_right_bound.y
            <= robot_extents.deck_extents.y + robot_extents.padding_rear
            and pip_front_right_bound.x >= robot_extents.padding_left_side
            and pip_back_right_bound.y >= robot_extents.padding_front
            and pip_back_right_bound.x >= robot_extents.padding_left_side
            and pip_front_left_bound.y
            <= robot_extents.deck_extents.y + robot_extents.padding_rear
            and pip_front_left_bound.x
            <= robot_extents.deck_extents.x + robot_extents.padding_right_side
            and pip_back_left_bound.y >= robot_extents.padding_front
            and pip_back_left_bound.x
            <= robot_extents.deck_extents.x + robot_extents.padding_right_side
        )
    # For 8ch pipettes we only check the rear and front extents
    return (
        pip_front_right_bound.y
        <= robot_extents.deck_extents.y + robot_extents.padding_rear
        and pip_back_right_bound.y >= robot_extents.padding_front
        and pip_front_left_bound.y
        <= robot_extents.deck_extents.y + robot_extents.padding_rear
        and pip_back_left_bound.y >= robot_extents.padding_front
    )



import logging
from typing import Dict, Optional

from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.pipette.pipette_load_name_conversions import (
    convert_to_pipette_name_type,
)
from opentrons_shared_data.pipette.types import PipetteGenerationType

from opentrons.types import Mount

from ..protocol import AbstractProtocol
from ..legacy.legacy_protocol_core import LegacyProtocolCore
from ..legacy.legacy_labware_core import LegacyLabwareCore
from ..legacy.legacy_module_core import LegacyModuleCore
from ..legacy.load_info import InstrumentLoadInfo

from .legacy_instrument_core import LegacyInstrumentCoreSimulator

logger = logging.getLogger(__name__)


class LegacyProtocolCoreSimulator(
    LegacyProtocolCore,
    AbstractProtocol[
        LegacyInstrumentCoreSimulator, LegacyLabwareCore, LegacyModuleCore
    ],
):
    _instruments: Dict[Mount, Optional[LegacyInstrumentCoreSimulator]]  # type: ignore[assignment]

    def load_instrument(  # type: ignore[override]
        self,
        instrument_name: PipetteNameType,
        mount: Mount,
        liquid_presence_detection: bool = False,
    ) -> LegacyInstrumentCoreSimulator:
        """Create a simulating instrument context."""
        pipette_generation = convert_to_pipette_name_type(
            instrument_name.value
        ).pipette_generation

        if pipette_generation not in [
            PipetteGenerationType.GEN1,
            PipetteGenerationType.GEN2,
        ]:
            raise ValueError(f"{instrument_name} is not a valid OT-2 pipette.")

        existing_instrument = self._instruments[mount]

        if (
            existing_instrument
            and existing_instrument.get_requested_as_name() == instrument_name.value
        ):
            # Replacing with the exact same instrument name. Just return the
            # existing instrument instance.
            return existing_instrument

        attached = {
            att_mount: instr.get("name", None)
            for att_mount, instr in self._sync_hardware.attached_instruments.items()
            if instr
        }
        attached[mount] = instrument_name.value
        self._sync_hardware.cache_instruments(attached)

        pipette_dict = self._sync_hardware.get_attached_instruments()[mount]
        new_instr = LegacyInstrumentCoreSimulator(
            protocol_interface=self,
            pipette_dict=pipette_dict,
            mount=mount,
            instrument_name=instrument_name.value,
            api_version=self._api_version,
        )
        self._instruments[mount] = new_instr

        logger.info(f"Instrument {new_instr} loaded")
        self._equipment_broker.publish(
            InstrumentLoadInfo(
                instrument_load_name=instrument_name.value,
                mount=mount,
                pipette_dict=pipette_dict,
            )
        )

        return new_instr



"""Protocol engine types for legacy labware offset locations.

This is its own module to fix circular imports.
"""

from typing import Optional, Literal, Annotated

from pydantic import BaseModel, Field

from opentrons.types import DeckSlotName

from .module import ModuleModel


class OnLabwareOffsetLocationSequenceComponent(BaseModel):
    """Offset location sequence component for a labware on another labware."""

    kind: Literal["onLabware"] = "onLabware"
    labwareUri: str = Field(
        ...,
        description=(
            "The definition URI of another labware, probably an adapter,"
            " that the labware will be loaded onto."
        ),
    )


class OnModuleOffsetLocationSequenceComponent(BaseModel):
    """Offset location sequence component for a labware on a module."""

    kind: Literal["onModule"] = "onModule"
    moduleModel: ModuleModel = Field(
        ..., description="The model of a module that a labware can be loaded on to."
    )


class OnAddressableAreaOffsetLocationSequenceComponent(BaseModel):
    """Offset location sequence component for a labware on an addressable area."""

    kind: Literal["onAddressableArea"] = "onAddressableArea"
    addressableAreaName: str = Field(
        ...,
        description=(
            'The ID of an addressable area that a labware or module can be loaded onto, such as (on the OT-2) "2" '
            'or (on the Flex) "C1". '
            "\n\n"
            "On the Flex, this field must be correct for the kind of entity it hosts. For instance, if the prior entity "
            "in the location sequence is an `OnModuleOffsetLocationSequenceComponent(moduleModel=temperatureModuleV2)`, "
            "this entity must be temperatureModuleV2NN where NN is the slot name in which the module resides. "
        ),
    )


LabwareOffsetLocationSequenceComponentsUnion = (
    OnLabwareOffsetLocationSequenceComponent
    | OnModuleOffsetLocationSequenceComponent
    | OnAddressableAreaOffsetLocationSequenceComponent
)

LabwareOffsetLocationSequenceComponents = Annotated[
    LabwareOffsetLocationSequenceComponentsUnion, Field(discriminator="kind")
]

LabwareOffsetLocationSequence = list[LabwareOffsetLocationSequenceComponents]


class LegacyLabwareOffsetLocation(BaseModel):
    """Parameters describing when a given offset may apply to a given labware load."""

    slotName: DeckSlotName = Field(
        ...,
        description=(
            "The deck slot where the protocol will load the labware."
            " Or, if the protocol will load the labware on a module,"
            " the deck slot where the protocol will load that module."
            "\n\n"
            # This description should be kept in sync with DeckSlotLocation.slotName.
            'The plain numbers like `"5"` are for the OT-2,'
            ' and the coordinates like `"C2"` are for the Flex.'
            "\n\n"
            "When you provide one of these values, you can use either style."
            " It will automatically be converted to match the robot."
            "\n\n"
            "When one of these values is returned, it will always match the robot."
        ),
    )
    moduleModel: Optional[ModuleModel] = Field(
        None,
        description=(
            "The model of the module that the labware will be loaded onto,"
            " if applicable."
            "\n\n"
            "Because of module compatibility, the model that the protocol requests"
            " may not be exactly the same"
            " as what it will find physically connected during execution."
            " For this labware offset to apply,"
            " this field must be the *requested* model, not the connected one."
            " You can retrieve this from a `loadModule` command's `params.model`"
            " in the protocol's analysis."
        ),
    )
    definitionUri: Optional[str] = Field(
        None,
        description=(
            "The definition URI of another labware, probably an adapter, that the"
            " labware will be loaded onto, if applicable."
            "\n\n"
            "This can be combined with moduleModel if the labware is loaded on top of"
            " an adapter that is loaded on a module."
        ),
    )



"""Move to coordinates command request, result, and implementation models."""
from __future__ import annotations

from pydantic import Field
from typing import Optional, Type, TYPE_CHECKING
from typing_extensions import Literal


from ..types import DeckPoint
from .pipetting_common import PipetteIdMixin
from .movement_common import (
    MovementMixin,
    DestinationPositionResult,
    move_to_coordinates,
    StallOrCollisionError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler
    from ..resources.model_utils import ModelUtils


MoveToCoordinatesCommandType = Literal["moveToCoordinates"]


class MoveToCoordinatesParams(PipetteIdMixin, MovementMixin):
    """Payload required to move a pipette to coordinates."""

    coordinates: DeckPoint = Field(
        ...,
        description="X, Y and Z coordinates in mm from deck's origin location (left-front-bottom corner of work space)",
    )


class MoveToCoordinatesResult(DestinationPositionResult):
    """Result data from the execution of a MoveToCoordinates command."""

    pass


_ExecuteReturn = (
    SuccessData[MoveToCoordinatesResult] | DefinedErrorData[StallOrCollisionError]
)


class MoveToCoordinatesImplementation(
    AbstractCommandImpl[MoveToCoordinatesParams, _ExecuteReturn]
):
    """Move to coordinates command implementation."""

    def __init__(
        self,
        movement: MovementHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._movement = movement
        self._model_utils = model_utils

    async def execute(self, params: MoveToCoordinatesParams) -> _ExecuteReturn:
        """Move the requested pipette to the requested coordinates."""
        result = await move_to_coordinates(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=params.pipetteId,
            deck_coordinates=params.coordinates,
            direct=params.forceDirect,
            additional_min_travel_z=params.minimumZHeight,
            speed=params.speed,
        )
        if isinstance(result, DefinedErrorData):
            return result
        else:
            return SuccessData(
                public=MoveToCoordinatesResult(position=result.public.position),
                state_update=result.state_update,
            )


class MoveToCoordinates(
    BaseCommand[MoveToCoordinatesParams, MoveToCoordinatesResult, StallOrCollisionError]
):
    """Move to well command model."""

    commandType: MoveToCoordinatesCommandType = "moveToCoordinates"
    params: MoveToCoordinatesParams
    result: Optional[MoveToCoordinatesResult] = None

    _ImplementationCls: Type[
        MoveToCoordinatesImplementation
    ] = MoveToCoordinatesImplementation


class MoveToCoordinatesCreate(BaseCommandCreate[MoveToCoordinatesParams]):
    """Move to coordinates command creation request model."""

    commandType: MoveToCoordinatesCommandType = "moveToCoordinates"
    params: MoveToCoordinatesParams

    _CommandCls: Type[MoveToCoordinates] = MoveToCoordinates



from typing import List
from opentrons.hardware_control.emulation.types import ModuleType
from opentrons.hardware_control.emulation.util import TEMPERATURE_ROOM
from pydantic import BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict


class PipetteSettings(BaseModel):
    model: str = "p20_single_v2.0"
    id: str = "P20SV202020070101"


class SmoothieSettings(BaseModel):
    left: PipetteSettings = PipetteSettings(
        model="p20_multi_v2.0", id="P3HMV202020041605"
    )
    right: PipetteSettings = PipetteSettings(
        model="p20_single_v2.0", id="P20SV202020070101"
    )
    host: str = "0.0.0.0"
    port: int = 9996


class BaseModuleSettings(BaseModel):
    serial_number: str
    model: str
    version: str


class TemperatureModelSettings(BaseModel):
    degrees_per_tick: float = 2.0
    starting: float = float(TEMPERATURE_ROOM)


class RPMModelSettings(BaseModel):
    rpm_per_tick: float = 100.0
    starting: float = 0.0


class MagDeckSettings(BaseModuleSettings):
    pass


class TempDeckSettings(BaseModuleSettings):
    temperature: TemperatureModelSettings


class ThermocyclerSettings(BaseModuleSettings):
    lid_temperature: TemperatureModelSettings
    plate_temperature: TemperatureModelSettings


class HeaterShakerSettings(BaseModuleSettings):
    temperature: TemperatureModelSettings
    rpm: RPMModelSettings
    home_delay_time: int = 0


class ProxySettings(BaseModel):
    """Settings for a proxy."""

    host: str = "0.0.0.0"
    emulator_port: int
    driver_port: int
    use_local_host: bool = True


class ModuleServerSettings(BaseModel):
    """Settings for the module server"""

    host: str = "0.0.0.0"
    port: int = 8989


class Settings(BaseSettings):
    modules: List[ModuleType] = [
        ModuleType.Magnetic,
        ModuleType.Temperature,
        ModuleType.Thermocycler,
    ]
    smoothie: SmoothieSettings = SmoothieSettings()
    magdeck: MagDeckSettings = MagDeckSettings(
        serial_number="magnetic_emulator", model="mag_deck_v20", version="2.0.0"
    )
    tempdeck: TempDeckSettings = TempDeckSettings(
        serial_number="temperature_emulator",
        model="temp_deck_v20",
        version="v2.0.1",
        temperature=TemperatureModelSettings(starting=0.0),
    )
    thermocycler: ThermocyclerSettings = ThermocyclerSettings(
        serial_number="thermocycler_emulator",
        model="v02",
        version="v1.1.0",
        lid_temperature=TemperatureModelSettings(),
        plate_temperature=TemperatureModelSettings(),
    )
    heatershaker: HeaterShakerSettings = HeaterShakerSettings(
        serial_number="heater_shaker_emulator",
        model="v01",
        version="v0.0.1",
        temperature=TemperatureModelSettings(),
        rpm=RPMModelSettings(),
        home_delay_time=0,
    )

    heatershaker_proxy: ProxySettings = ProxySettings(
        emulator_port=9000, driver_port=9995
    )
    thermocycler_proxy: ProxySettings = ProxySettings(
        emulator_port=9002, driver_port=9997
    )
    temperature_proxy: ProxySettings = ProxySettings(
        emulator_port=9003, driver_port=9998
    )
    magdeck_proxy: ProxySettings = ProxySettings(emulator_port=9004, driver_port=9999)
    model_config = SettingsConfigDict(env_prefix="OT_EMULATOR_")

    module_server: ModuleServerSettings = ModuleServerSettings()



"""setStatusBar command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal
import enum

from opentrons.hardware_control.types import StatusBarState
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ..execution import StatusBarHandler

SetStatusBarCommandType = Literal["setStatusBar"]


class StatusBarAnimation(enum.Enum):
    """Status Bar animation options."""

    IDLE = "idle"
    CONFIRM = "confirm"
    UPDATING = "updating"
    DISCO = "disco"
    OFF = "off"


def _animation_to_status_bar_state(animation: StatusBarAnimation) -> StatusBarState:
    return {
        StatusBarAnimation.IDLE: StatusBarState.IDLE,
        StatusBarAnimation.CONFIRM: StatusBarState.CONFIRMATION,
        StatusBarAnimation.UPDATING: StatusBarState.UPDATING,
        StatusBarAnimation.DISCO: StatusBarState.DISCO,
        StatusBarAnimation.OFF: StatusBarState.OFF,
    }[animation]


class SetStatusBarParams(BaseModel):
    """Payload required to set the status bar to run an animation."""

    animation: StatusBarAnimation = Field(
        ...,
        description="The animation that should be executed on the status bar.",
    )


class SetStatusBarResult(BaseModel):
    """Result data from the execution of a SetStatusBar command."""


class SetStatusBarImplementation(
    AbstractCommandImpl[SetStatusBarParams, SuccessData[SetStatusBarResult]]
):
    """setStatusBar command implementation."""

    def __init__(self, status_bar: StatusBarHandler, **kwargs: object) -> None:
        self._status_bar = status_bar

    async def execute(
        self, params: SetStatusBarParams
    ) -> SuccessData[SetStatusBarResult]:
        """Execute the setStatusBar command."""
        if not self._status_bar.status_bar_should_not_be_changed():
            state = _animation_to_status_bar_state(params.animation)
            await self._status_bar.set_status_bar(state)
        return SuccessData(
            public=SetStatusBarResult(),
        )


class SetStatusBar(
    BaseCommand[SetStatusBarParams, SetStatusBarResult, ErrorOccurrence]
):
    """setStatusBar command model."""

    commandType: SetStatusBarCommandType = "setStatusBar"
    params: SetStatusBarParams
    result: Optional[SetStatusBarResult] = None

    _ImplementationCls: Type[SetStatusBarImplementation] = SetStatusBarImplementation


class SetStatusBarCreate(BaseCommandCreate[SetStatusBarParams]):
    """setStatusBar command request model."""

    commandType: SetStatusBarCommandType = "setStatusBar"
    params: SetStatusBarParams

    _CommandCls: Type[SetStatusBar] = SetStatusBar



from __future__ import annotations

import asyncio
from typing import Dict, Optional, List, Tuple, TYPE_CHECKING

from opentrons.drivers.types import (
    ABSMeasurementMode,
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderDeviceState,
    AbsorbanceReaderPlatePresence,
)
from opentrons.drivers.absorbance_reader.abstract import (
    AbstractAbsorbanceReaderDriver,
)
from opentrons.drivers.rpi_drivers.types import USBPort

if TYPE_CHECKING:
    from .async_byonoy import AsyncByonoy as AsyncByonoyType


class AbsorbanceReaderDriver(AbstractAbsorbanceReaderDriver):
    @classmethod
    async def create(
        cls,
        port: str,
        usb_port: USBPort,
        loop: Optional[asyncio.AbstractEventLoop],
    ) -> AbsorbanceReaderDriver:
        """Create an absorbance reader driver."""
        from .async_byonoy import AsyncByonoy

        connection = await AsyncByonoy.create(port=port, usb_port=usb_port, loop=loop)
        return cls(connection=connection)

    def __init__(self, connection: AsyncByonoyType) -> None:
        self._connection = connection

    async def connect(self) -> None:
        """Connect to absorbance reader"""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from absorbance reader"""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check connection to absorbance reader"""
        return await self._connection.is_open()

    async def get_device_info(self) -> Dict[str, str]:
        """Get device info"""
        return await self._connection.get_device_information()

    async def update_firmware(self, firmware_file_path: str) -> Tuple[bool, str]:
        return await self._connection.update_firmware(firmware_file_path)

    async def get_uptime(self) -> int:
        return await self._connection.get_device_uptime()

    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        return await self._connection.get_lid_status()

    async def get_available_wavelengths(self) -> List[int]:
        return await self._connection.get_supported_wavelengths()

    async def initialize_measurement(
        self,
        wavelengths: List[int],
        mode: ABSMeasurementMode = ABSMeasurementMode.SINGLE,
        reference_wavelength: Optional[int] = None,
    ) -> None:
        await self._connection.initialize(mode, wavelengths, reference_wavelength)

    async def get_measurement(self) -> List[List[float]]:
        return await self._connection.get_measurement()

    async def get_status(self) -> AbsorbanceReaderDeviceState:
        return await self._connection.get_device_status()

    async def get_plate_presence(self) -> AbsorbanceReaderPlatePresence:
        return await self._connection.get_plate_presence()



"""Update position estimators payload, result, and implementaiton."""

from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, List, Type
from typing_extensions import Literal

from ...types import MotorAxis
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...resources import ensure_ot3_hardware

from opentrons.hardware_control import HardwareControlAPI

if TYPE_CHECKING:
    from ...execution import GantryMover


UpdatePositionEstimatorsCommandType = Literal["unsafe/updatePositionEstimators"]


class UpdatePositionEstimatorsParams(BaseModel):
    """Payload required for an UpdatePositionEstimators command."""

    axes: List[MotorAxis] = Field(
        ...,
        description=(
            "The axes for which to update the position estimators."
            " Any axes that are not physically present will be ignored."
        ),
    )


class UpdatePositionEstimatorsResult(BaseModel):
    """Result data from the execution of an UpdatePositionEstimators command."""


class UpdatePositionEstimatorsImplementation(
    AbstractCommandImpl[
        UpdatePositionEstimatorsParams,
        SuccessData[UpdatePositionEstimatorsResult],
    ]
):
    """Update position estimators command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api
        self._gantry_mover = gantry_mover

    async def execute(
        self, params: UpdatePositionEstimatorsParams
    ) -> SuccessData[UpdatePositionEstimatorsResult]:
        """Update axis position estimators from their encoders."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        await ot3_hardware_api.update_axis_position_estimations(
            self._gantry_mover.motor_axes_to_present_hardware_axes(params.axes)
        )
        return SuccessData(
            public=UpdatePositionEstimatorsResult(),
        )


class UpdatePositionEstimators(
    BaseCommand[
        UpdatePositionEstimatorsParams, UpdatePositionEstimatorsResult, ErrorOccurrence
    ]
):
    """UpdatePositionEstimators command model."""

    commandType: UpdatePositionEstimatorsCommandType = "unsafe/updatePositionEstimators"
    params: UpdatePositionEstimatorsParams
    result: Optional[UpdatePositionEstimatorsResult] = None

    _ImplementationCls: Type[
        UpdatePositionEstimatorsImplementation
    ] = UpdatePositionEstimatorsImplementation


class UpdatePositionEstimatorsCreate(BaseCommandCreate[UpdatePositionEstimatorsParams]):
    """UpdatePositionEstimators command request model."""

    commandType: UpdatePositionEstimatorsCommandType = "unsafe/updatePositionEstimators"
    params: UpdatePositionEstimatorsParams

    _CommandCls: Type[UpdatePositionEstimators] = UpdatePositionEstimators



""" opentrons.protocols: functions and state for handling, executing, and
simulating protocols.

This genericizes over v1 and v2 protocols and allows for version inference
and version-independent apis for simulating and executing protocols.
"""



"""Calibration commands."""
from .calibrate_gripper import (
    CalibrateGripperParams,
    CalibrateGripperResult,
    CalibrateGripperCreate,
    CalibrateGripperCommandType,
    CalibrateGripper,
)

from .calibrate_pipette import (
    CalibratePipetteParams,
    CalibratePipetteResult,
    CalibratePipetteCreate,
    CalibratePipetteCommandType,
    CalibratePipette,
)

from .calibrate_module import (
    CalibrateModuleParams,
    CalibrateModuleResult,
    CalibrateModuleCreate,
    CalibrateModuleCommandType,
    CalibrateModule,
)

from .move_to_maintenance_position import (
    MoveToMaintenancePositionParams,
    MoveToMaintenancePositionResult,
    MoveToMaintenancePositionCreate,
    MoveToMaintenancePositionCommandType,
    MoveToMaintenancePosition,
)

__all__ = [
    # calibration/calibrateModule
    "CalibrateModule",
    "CalibrateModuleCreate",
    "CalibrateModuleParams",
    "CalibrateModuleResult",
    "CalibrateModuleCommandType",
    # calibration/calibrateGripper
    "CalibrateGripper",
    "CalibrateGripperCreate",
    "CalibrateGripperParams",
    "CalibrateGripperParamsProbe",
    "CalibrateGripperResult",
    "CalibrateGripperCommandType",
    # calibration/calibratePipette
    "CalibratePipette",
    "CalibratePipetteCreate",
    "CalibratePipetteParams",
    "CalibratePipetteResult",
    "CalibratePipetteCommandType",
    # calibration/moveToMaintenancePosition
    "MoveToMaintenancePosition",
    "MoveToMaintenancePositionCreate",
    "MoveToMaintenancePositionParams",
    "MoveToMaintenancePositionResult",
    "MoveToMaintenancePositionCommandType",
]



from __future__ import annotations

import asyncio
import logging
from typing import Dict, Optional, Mapping

from opentrons.drivers.flex_stacker.types import (
    Direction,
    LEDColor,
    LEDPattern,
    MoveParams,
    MoveResult,
    StackerAxis,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.flex_stacker.driver import (
    STACKER_MOTION_CONFIG,
    STALLGUARD_CONFIG,
    FlexStackerDriver,
)
from opentrons.drivers.flex_stacker.abstract import AbstractFlexStackerDriver
from opentrons.drivers.flex_stacker.simulator import SimulatingDriver
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.poller import Reader, Poller
from opentrons.hardware_control.modules import mod_abc, update
from opentrons.hardware_control.modules.types import (
    FlexStackerStatus,
    HopperDoorState,
    LatchState,
    ModuleDisconnectedCallback,
    ModuleType,
    PlatformState,
    StackerAxisState,
    UploadFunction,
    LiveData,
    FlexStackerData,
)

from opentrons_shared_data.errors.exceptions import FlexStackerStallError

log = logging.getLogger(__name__)

POLL_PERIOD = 1.0
SIMULATING_POLL_PERIOD = POLL_PERIOD / 20.0

DFU_PID = "df11"

# Maximum distance in mm the axis can travel.
MAX_TRAVEL = {
    StackerAxis.X: 194.0,
    StackerAxis.Z: 139.5,
    StackerAxis.L: 22.0,
}

# The offset in mm to subtract from MAX_TRAVEL when moving an axis before we home.
# This lets us use `move_axis` to move fast, leaving the axis OFFSET mm
# from the limit switch. Then we can use `home_axis` to move the axis the rest
# of the way until we trigger the expected limit switch.
OFFSET_SM = 5.0
OFFSET_MD = 10.0
OFFSET_LG = 20.0

# height limit in mm of labware to use OFFSET_MD used when storing labware.
MEDIUM_LABWARE_Z_LIMIT = 20.0


class FlexStacker(mod_abc.AbstractModule):
    """Hardware control interface for an attached Flex-Stacker module."""

    MODULE_TYPE = ModuleType.FLEX_STACKER

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> "FlexStacker":
        """
        Build a FlexStacker

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            FlexStacker instance
        """
        driver: AbstractFlexStackerDriver
        if not simulating:
            driver = await FlexStackerDriver.create(port=port, loop=hw_control_loop)
            poll_interval_seconds = poll_interval_seconds or POLL_PERIOD
        else:
            driver = SimulatingDriver(serial_number=sim_serial_number)
            poll_interval_seconds = poll_interval_seconds or SIMULATING_POLL_PERIOD

        reader = FlexStackerReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=(await driver.get_device_info()).to_dict(),
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )

        # Enable stallguard
        for axis, config in STALLGUARD_CONFIG.items():
            await driver.set_stallguard_threshold(
                axis, config.enabled, config.threshold
            )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of Flex-Stacker on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractFlexStackerDriver,
        reader: FlexStackerReader,
        poller: Poller,
        device_info: Mapping[str, str],
        hw_control_loop: asyncio.AbstractEventLoop,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ):
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )
        self._device_info = device_info
        self._driver = driver
        self._reader = reader
        self._poller = poller
        self._stacker_status = FlexStackerStatus.IDLE
        self._stall_detected = False

    async def cleanup(self) -> None:
        """Stop the poller task"""
        await self._poller.stop()
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        """Used for picking up serial port symlinks"""
        return "flexstacker"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "flex-stacker"

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        return "flexStackerModuleV1"

    def model(self) -> str:
        return self._model_from_revision(self._device_info.get("model"))

    @property
    def latch_state(self) -> LatchState:
        """The state of the latch."""
        return LatchState.from_state(self.limit_switch_status[StackerAxis.L])

    @property
    def platform_state(self) -> PlatformState:
        """The state of the platform."""
        return self._reader.platform_state

    @property
    def hopper_door_state(self) -> HopperDoorState:
        """The status of the hopper door."""
        return HopperDoorState.from_state(self._reader.hopper_door_closed)

    @property
    def limit_switch_status(self) -> Dict[StackerAxis, StackerAxisState]:
        """The status of the Limit switches."""
        return self._reader.limit_switch_status

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def status(self) -> FlexStackerStatus:
        """Module status or error state details."""
        return self._stacker_status

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    @property
    def live_data(self) -> LiveData:
        data: FlexStackerData = {
            "latchState": self.latch_state.value,
            "platformState": self.platform_state.value,
            "hopperDoorState": self.hopper_door_state.value,
            "axisStateX": self.limit_switch_status[StackerAxis.X].value,
            "axisStateZ": self.limit_switch_status[StackerAxis.Z].value,
            "errorDetails": self._reader.error,
        }
        return {"status": self.status.value, "data": data}

    async def prep_for_update(self) -> str:
        await self._poller.stop()
        await self._driver.stop_motors()
        await self._driver.enter_programming_mode()
        # flex stacker has three unique "devices" over DFU
        dfu_info = await update.find_dfu_device(pid=DFU_PID, expected_device_count=3)
        return dfu_info

    def bootloader(self) -> UploadFunction:
        return update.upload_via_dfu

    async def deactivate(self, must_be_running: bool = True) -> None:
        await self._driver.stop_motors()

    async def reset_stall_detected(self) -> None:
        """Sets the statusbar to normal."""
        if self._stall_detected:
            await self.set_led_state(0.5, LEDColor.GREEN, LEDPattern.STATIC)
            self._stall_detected = False

    async def set_led_state(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,
        reps: Optional[int] = None,
    ) -> None:
        """Sets the statusbar state."""
        return await self._driver.set_led(
            power, color=color, pattern=pattern, duration=duration, reps=reps
        )

    async def move_axis(
        self,
        axis: StackerAxis,
        direction: Direction,
        distance: float,
        speed: Optional[float] = None,
        acceleration: Optional[float] = None,
        current: Optional[float] = None,
    ) -> bool:
        """Move the axis in a direction by the given distance in mm."""
        await self.reset_stall_detected()
        default = STACKER_MOTION_CONFIG[axis]["move"]
        await self._driver.set_run_current(
            axis, current if current is not None else default.run_current
        )
        await self._driver.set_ihold_current(axis, default.hold_current)
        motion_params = default.move_params.update(
            max_speed=speed, acceleration=acceleration
        )
        distance = direction.distance(distance)
        res = await self._driver.move_in_mm(axis, distance, params=motion_params)
        if res == MoveResult.STALL_ERROR:
            self._stall_detected = True
            raise FlexStackerStallError(self.device_info["serial"], axis)
        return res == MoveResult.NO_ERROR

    async def home_axis(
        self,
        axis: StackerAxis,
        direction: Direction,
        speed: Optional[float] = None,
        acceleration: Optional[float] = None,
        current: Optional[float] = None,
    ) -> bool:
        await self.reset_stall_detected()
        default = STACKER_MOTION_CONFIG[axis]["home"]
        await self._driver.set_run_current(
            axis, current if current is not None else default.run_current
        )
        await self._driver.set_ihold_current(axis, default.hold_current)
        motion_params = default.move_params.update(
            max_speed=speed, acceleration=acceleration
        )
        success = await self._driver.move_to_limit_switch(
            axis=axis, direction=direction, params=motion_params
        )
        if success == MoveResult.STALL_ERROR:
            self._stall_detected = True
            raise FlexStackerStallError(self.device_info["serial"], axis)
        return success == MoveResult.NO_ERROR

    async def close_latch(
        self,
        velocity: Optional[float] = None,
        acceleration: Optional[float] = None,
    ) -> bool:
        """Close the latch, dropping any labware its holding."""
        # Dont move the latch if its already closed.
        if self.limit_switch_status[StackerAxis.L] == StackerAxisState.EXTENDED:
            return True
        success = await self.home_axis(
            StackerAxis.L,
            Direction.RETRACT,
            speed=velocity,
            acceleration=acceleration,
        )
        # Check that the latch is closed.
        await self._reader.get_limit_switch_status()
        return (
            success
            and self.limit_switch_status[StackerAxis.L] == StackerAxisState.EXTENDED
        )

    async def open_latch(
        self,
        velocity: Optional[float] = None,
        acceleration: Optional[float] = None,
    ) -> bool:
        """Open the latch."""
        # Dont move the latch if its already opened.
        if self.limit_switch_status[StackerAxis.L] == StackerAxisState.RETRACTED:
            return True
        # The latch only has one limit switch, so we have to travel a fixed distance
        # to open the latch.
        success = await self.move_axis(
            StackerAxis.L,
            Direction.EXTEND,
            distance=MAX_TRAVEL[StackerAxis.L],
            speed=velocity,
            acceleration=acceleration,
        )
        # Check that the latch is opened.
        await self._reader.get_limit_switch_status()
        axis_state = self.limit_switch_status[StackerAxis.L]
        return success and axis_state == StackerAxisState.RETRACTED

    async def dispense_labware(self, labware_height: float) -> bool:
        """Dispenses the next labware in the stacker."""
        await self._prepare_for_action()

        # Move platform along the X then Z axis
        await self._move_and_home_axis(StackerAxis.X, Direction.RETRACT, OFFSET_SM)
        await self._move_and_home_axis(StackerAxis.Z, Direction.EXTEND, OFFSET_SM)

        # Transfer
        await self.open_latch()
        await self.move_axis(StackerAxis.Z, Direction.RETRACT, (labware_height / 2) + 2)
        await self.close_latch()

        # Move platform along the Z then X axis
        offset = labware_height / 2 + OFFSET_MD
        await self._move_and_home_axis(StackerAxis.Z, Direction.RETRACT, offset)
        await self._move_and_home_axis(StackerAxis.X, Direction.EXTEND, OFFSET_SM)
        return True

    async def store_labware(self, labware_height: float) -> bool:
        """Stores a labware in the stacker."""
        await self._prepare_for_action()

        # Move X then Z axis
        offset = OFFSET_MD if labware_height < MEDIUM_LABWARE_Z_LIMIT else OFFSET_LG * 2
        distance = MAX_TRAVEL[StackerAxis.Z] - (labware_height / 2) - offset
        await self._move_and_home_axis(StackerAxis.X, Direction.RETRACT, OFFSET_SM)
        await self.move_axis(StackerAxis.Z, Direction.EXTEND, distance)

        # Transfer
        await self.open_latch()
        z_speed = (
            STACKER_MOTION_CONFIG[StackerAxis.Z]["move"].move_params.max_speed or 0
        ) / 2
        await self.move_axis(
            StackerAxis.Z, Direction.EXTEND, (labware_height / 2), z_speed
        )
        await self.home_axis(StackerAxis.Z, Direction.EXTEND, z_speed)
        await self.close_latch()

        # Move Z then X axis
        await self._move_and_home_axis(StackerAxis.Z, Direction.RETRACT, OFFSET_LG)
        await self._move_and_home_axis(StackerAxis.X, Direction.EXTEND, OFFSET_SM)
        return True

    async def _move_and_home_axis(
        self, axis: StackerAxis, direction: Direction, offset: float = 0
    ) -> bool:
        distance = MAX_TRAVEL[axis] - offset
        await self.move_axis(axis, direction, distance)
        return await self.home_axis(axis, direction)

    async def _prepare_for_action(self) -> bool:
        """Helper to prepare axis for dispensing or storing labware."""
        # TODO: check if we need to home first
        await self.home_axis(StackerAxis.X, Direction.EXTEND)
        await self.home_axis(StackerAxis.Z, Direction.RETRACT)
        await self.close_latch()
        return True

    async def home_all(self, ignore_latch: bool = False) -> None:
        """Home all axes based on current state, assuming normal operation.

        If ignore_latch is True, we will not attempt to close the latch. This
        is useful when we want the shuttle to be out of the way for error
        recovery (e.g. when the latch is stuck open).
        """
        await self._reader.read()
        # we should always be able to home the X axis first
        await self.home_axis(StackerAxis.X, Direction.RETRACT)
        # If latch is open, we must first close it
        if not ignore_latch and self.latch_state == LatchState.OPENED:
            if self.limit_switch_status[StackerAxis.Z] != StackerAxisState.RETRACTED:
                # it was likely in the middle of a dispense/store command
                # z should be moved up before we can safely close the latch
                await self.home_axis(StackerAxis.Z, Direction.EXTEND)
            await self.close_latch()
        await self.home_axis(StackerAxis.Z, Direction.RETRACT)
        await self.home_axis(StackerAxis.X, Direction.EXTEND)


class FlexStackerReader(Reader):
    error: Optional[str]

    def __init__(self, driver: AbstractFlexStackerDriver) -> None:
        self.error: Optional[str] = None
        self._driver = driver
        self.limit_switch_status = {
            axis: StackerAxisState.UNKNOWN for axis in StackerAxis
        }
        self.platform_state = PlatformState.UNKNOWN
        self.hopper_door_closed = False
        self.motion_params: Dict[StackerAxis, Optional[MoveParams]] = {
            axis: None for axis in StackerAxis
        }
        self.get_config = True

    async def read(self) -> None:
        await self.get_limit_switch_status()
        await self.get_platform_sensor_state()
        await self.get_door_closed()
        if self.get_config:
            await self.get_motion_parameters()
            self.get_config = False
        self._set_error(None)

    async def get_limit_switch_status(self) -> None:
        """Get the limit switch status."""
        status = await self._driver.get_limit_switches_status()
        self.limit_switch_status = {
            axis: StackerAxisState.from_status(status, axis) for axis in StackerAxis
        }

    async def get_motion_parameters(self) -> None:
        """Get the motion parameters used by the axis motors."""
        self.motion_params = {
            axis: await self._driver.get_motion_params(axis) for axis in StackerAxis
        }

    async def get_platform_sensor_state(self) -> None:
        """Get the platform state."""
        status = await self._driver.get_platform_status()
        self.platform_state = PlatformState.from_status(status)

    async def get_door_closed(self) -> None:
        """Check if the hopper door is closed."""
        self.hopper_door_closed = await self._driver.get_hopper_door_closed()

    def on_error(self, exception: Exception) -> None:
        self._driver.reset_serial_buffers()
        self._set_error(exception)

    def _set_error(self, exception: Optional[Exception]) -> None:
        if exception is None:
            self.error = None
        else:
            try:
                self.error = str(exception.args[0])
            except Exception:
                self.error = repr(exception)



from __future__ import annotations
from typing import NamedTuple, TypedDict


class APIVersion(NamedTuple):
    major: int
    minor: int

    @classmethod
    def from_string(cls, inp: str) -> APIVersion:
        parts = inp.split(".")
        if len(parts) != 2:
            raise ValueError(inp)
        intparts = [int(p) for p in parts]

        return cls(major=intparts[0], minor=intparts[1])

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"


class ThermocyclerStepBase(TypedDict):
    """Required elements of a thermocycler step: the temperature."""

    temperature: float


class ThermocyclerStep(ThermocyclerStepBase, total=False):
    """Optional elements of a thermocycler step: the hold time. One of these must be present."""

    hold_time_seconds: float
    hold_time_minutes: float



"""Flex Stacker substate."""

from dataclasses import dataclass
from typing import NewType
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons.protocol_engine.state.update_types import (
    FlexStackerStateUpdate,
    NO_CHANGE,
)


FlexStackerId = NewType("FlexStackerId", str)


@dataclass(frozen=True)
class FlexStackerSubState:
    """Flex Stacker-specific state.

    Provides calculations and read-only state access
    for an individual loaded Flex Stacker Module.
    """

    module_id: FlexStackerId
    pool_primary_definition: LabwareDefinition | None
    pool_adapter_definition: LabwareDefinition | None
    pool_lid_definition: LabwareDefinition | None
    pool_count: int
    max_pool_count: int

    def new_from_state_change(
        self, update: FlexStackerStateUpdate
    ) -> "FlexStackerSubState":
        """Return a new state with the given update applied."""
        pool_primary_definition = self.pool_primary_definition
        pool_adapter_definition = self.pool_adapter_definition
        pool_lid_definition = self.pool_lid_definition
        max_pool_count = self.max_pool_count
        if update.pool_constraint != NO_CHANGE:
            max_pool_count = update.pool_constraint.max_pool_count
            pool_primary_definition = update.pool_constraint.primary_definition
            pool_adapter_definition = update.pool_constraint.adapter_definition
            pool_lid_definition = update.pool_constraint.lid_definition

        pool_count = self.pool_count
        if update.pool_count != NO_CHANGE:
            pool_count = update.pool_count

        return FlexStackerSubState(
            module_id=self.module_id,
            pool_primary_definition=pool_primary_definition,
            pool_adapter_definition=pool_adapter_definition,
            pool_lid_definition=pool_lid_definition,
            pool_count=pool_count,
            max_pool_count=max_pool_count,
        )

    def get_pool_definition_ordered_list(self) -> list[LabwareDefinition] | None:
        """Get the pool definitions in a list suitable for getting the height."""
        if not self.pool_primary_definition:
            return None

        defs: list[LabwareDefinition] = []
        if self.pool_lid_definition is not None:
            defs.append(self.pool_lid_definition)
        defs.append(self.pool_primary_definition)
        if self.pool_adapter_definition is not None:
            defs.append(self.pool_adapter_definition)
        return defs



"""Tip state tracking."""

from dataclasses import dataclass
from enum import Enum
from typing import Dict, Iterable, Optional, List, Union

from opentrons.types import NozzleMapInterface
from opentrons.protocol_engine.state import update_types

from ._abstract_store import HasState, HandlesActions
from ._well_math import wells_covered_dense
from ..actions import Action, ResetTipsAction, get_state_updates

from opentrons.hardware_control.nozzle_manager import NozzleMap


class _TipRackWellState(Enum):
    """The state of a single tip in a tip rack's well."""

    CLEAN = "clean"
    USED = "used"


_TipRackStateByWellName = Dict[str, _TipRackWellState]


@dataclass
class TipState:
    """State of all tips."""

    tips_by_labware_id: Dict[str, _TipRackStateByWellName]
    columns_by_labware_id: Dict[str, List[List[str]]]


class TipStore(HasState[TipState], HandlesActions):
    """Tip state container."""

    _state: TipState

    def __init__(self) -> None:
        """Initialize a liquid store and its state."""
        self._state = TipState(
            tips_by_labware_id={},
            columns_by_labware_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._handle_state_update(state_update)

        if isinstance(action, ResetTipsAction):
            labware_id = action.labware_id

            for well_name in self._state.tips_by_labware_id[labware_id].keys():
                self._state.tips_by_labware_id[labware_id][
                    well_name
                ] = _TipRackWellState.CLEAN

    def _handle_state_update(self, state_update: update_types.StateUpdate) -> None:
        if state_update.tips_used != update_types.NO_CHANGE:
            self._set_used_tips(
                labware_id=state_update.tips_used.labware_id,
                well_names=state_update.tips_used.well_names,
            )

        if state_update.loaded_labware != update_types.NO_CHANGE:
            labware_id = state_update.loaded_labware.labware_id
            definition = state_update.loaded_labware.definition
            if definition.parameters.isTiprack:
                self._state.tips_by_labware_id[labware_id] = {
                    well_name: _TipRackWellState.CLEAN
                    for column in definition.ordering
                    for well_name in column
                }
                self._state.columns_by_labware_id[labware_id] = [
                    column for column in definition.ordering
                ]
        if state_update.batch_loaded_labware != update_types.NO_CHANGE:
            for labware_id in state_update.batch_loaded_labware.new_locations_by_id:
                definition = state_update.batch_loaded_labware.definitions_by_id[
                    labware_id
                ]
                if definition.parameters.isTiprack:
                    self._state.tips_by_labware_id[labware_id] = {
                        well_name: _TipRackWellState.CLEAN
                        for column in definition.ordering
                        for well_name in column
                    }
                    self._state.columns_by_labware_id[labware_id] = [
                        column for column in definition.ordering
                    ]

    def _set_used_tips(self, labware_id: str, well_names: Iterable[str]) -> None:
        well_states = self._state.tips_by_labware_id.get(labware_id, {})
        for well_name in well_names:
            well_states[well_name] = _TipRackWellState.USED


class TipView:
    """Read-only tip state view."""

    _state: TipState

    def __init__(self, state: TipState) -> None:
        """Initialize the computed view of liquid state.

        Arguments:
            state: Liquid state dataclass used for all calculations.
        """
        self._state = state

    def get_next_tip(  # noqa: C901
        self,
        labware_id: str,
        num_tips: int,
        starting_tip_name: Optional[str],
        nozzle_map: Optional[NozzleMapInterface],
    ) -> Optional[str]:
        """Get the next available clean tip. Does not support use of a starting tip if the pipette used is in a partial configuration."""
        wells = self._state.tips_by_labware_id.get(labware_id, {})
        columns = self._state.columns_by_labware_id.get(labware_id, [])

        # TODO(sf): I'm pretty sure this can be replaced with wells_covered_96 but I'm not quite sure how
        def _identify_tip_cluster(
            active_columns: int,
            active_rows: int,
            critical_column: int,
            critical_row: int,
            entry_well: str,
        ) -> Optional[List[str]]:
            tip_cluster: list[str] = []
            for i in range(active_columns):
                if entry_well == "A1" or entry_well == "H1":
                    if critical_column - i >= 0:
                        column = columns[critical_column - i]
                    else:
                        return None
                elif entry_well == "A12" or entry_well == "H12":
                    if critical_column + i < len(columns):
                        column = columns[critical_column + i]
                    else:
                        return None
                else:
                    raise ValueError(
                        f"Invalid entry well {entry_well} for tip cluster identification."
                    )
                for j in range(active_rows):
                    if entry_well == "A1" or entry_well == "A12":
                        if critical_row - j >= 0:
                            well = column[critical_row - j]
                        else:
                            return None
                    elif entry_well == "H1" or entry_well == "H12":
                        if critical_row + j < len(column):
                            well = column[critical_row + j]
                        else:
                            return None
                    tip_cluster.append(well)

            if any(well not in [*wells] for well in tip_cluster):
                return None

            return tip_cluster

        def _validate_tip_cluster(
            active_columns: int, active_rows: int, tip_cluster: List[str]
        ) -> Union[str, int, None]:
            if not any(wells[well] == _TipRackWellState.USED for well in tip_cluster):
                return tip_cluster[0]
            elif all(wells[well] == _TipRackWellState.USED for well in tip_cluster):
                return None
            else:
                # In the case of an 8ch pipette where a column has mixed state tips we may simply progress to the next column in our search
                if nozzle_map is not None and nozzle_map.physical_nozzle_count == 8:
                    return None

                # In the case of a 96ch we can attempt to index in by singular rows and columns assuming that indexed direction is safe
                # The tip cluster list is ordered: Each row from a column in order by columns
                tip_cluster_final_column: list[str] = []
                for i in range(active_rows):
                    tip_cluster_final_column.append(
                        tip_cluster[((active_columns * active_rows) - 1) - i]
                    )
                tip_cluster_final_row: list[str] = []
                for i in range(active_columns):
                    tip_cluster_final_row.append(
                        tip_cluster[(active_rows - 1) + (i * active_rows)]
                    )
                if all(
                    wells[well] == _TipRackWellState.USED
                    for well in tip_cluster_final_column
                ):
                    return None
                elif all(
                    wells[well] == _TipRackWellState.USED
                    for well in tip_cluster_final_row
                ):
                    return None
                else:
                    # Tiprack has no valid tip selection, cannot progress
                    return -1

        # Search through the tiprack beginning at A1
        def _cluster_search_A1(active_columns: int, active_rows: int) -> Optional[str]:
            critical_column = active_columns - 1
            critical_row = active_rows - 1

            while critical_column < len(columns):
                tip_cluster = _identify_tip_cluster(
                    active_columns, active_rows, critical_column, critical_row, "A1"
                )
                if tip_cluster is not None:
                    result = _validate_tip_cluster(
                        active_columns, active_rows, tip_cluster
                    )
                    if isinstance(result, str):
                        return result
                    elif isinstance(result, int) and result == -1:
                        return None
                if critical_row + 1 < len(columns[0]):
                    critical_row = critical_row + 1
                else:
                    critical_column += 1
                    critical_row = active_rows - 1
            return None

        # Search through the tiprack beginning at A12
        def _cluster_search_A12(active_columns: int, active_rows: int) -> Optional[str]:
            critical_column = len(columns) - active_columns
            critical_row = active_rows - 1

            while critical_column >= 0:
                tip_cluster = _identify_tip_cluster(
                    active_columns, active_rows, critical_column, critical_row, "A12"
                )
                if tip_cluster is not None:
                    result = _validate_tip_cluster(
                        active_columns, active_rows, tip_cluster
                    )
                    if isinstance(result, str):
                        return result
                    elif isinstance(result, int) and result == -1:
                        return None
                if critical_row + 1 < len(columns[0]):
                    critical_row = critical_row + 1
                else:
                    critical_column -= 1
                    critical_row = active_rows - 1
            return None

        # Search through the tiprack beginning at H1
        def _cluster_search_H1(active_columns: int, active_rows: int) -> Optional[str]:
            critical_column = active_columns - 1
            critical_row = len(columns[critical_column]) - active_rows

            while critical_column <= len(columns):  # change to max size of labware
                tip_cluster = _identify_tip_cluster(
                    active_columns, active_rows, critical_column, critical_row, "H1"
                )
                if tip_cluster is not None:
                    result = _validate_tip_cluster(
                        active_columns, active_rows, tip_cluster
                    )
                    if isinstance(result, str):
                        return result
                    elif isinstance(result, int) and result == -1:
                        return None
                if critical_row - 1 >= 0:
                    critical_row = critical_row - 1
                else:
                    critical_column += 1
                    if critical_column >= len(columns):
                        return None
                    critical_row = len(columns[critical_column]) - active_rows
            return None

        # Search through the tiprack beginning at H12
        def _cluster_search_H12(active_columns: int, active_rows: int) -> Optional[str]:
            critical_column = len(columns) - active_columns
            critical_row = len(columns[critical_column]) - active_rows

            while critical_column >= 0:
                tip_cluster = _identify_tip_cluster(
                    active_columns, active_rows, critical_column, critical_row, "H12"
                )
                if tip_cluster is not None:
                    result = _validate_tip_cluster(
                        active_columns, active_rows, tip_cluster
                    )
                    if isinstance(result, str):
                        return result
                    elif isinstance(result, int) and result == -1:
                        return None
                if critical_row - 1 >= 0:
                    critical_row = critical_row - 1
                else:
                    critical_column -= 1
                    if critical_column < 0:
                        return None
                    critical_row = len(columns[critical_column]) - active_rows
            return None

        if starting_tip_name is None and nozzle_map is not None and columns:
            num_channels = nozzle_map.physical_nozzle_count
            num_nozzle_cols = len(nozzle_map.columns)
            num_nozzle_rows = len(nozzle_map.rows)
            # Each pipette's cluster search is determined by the point of entry for a given pipette/configuration:
            # - Single channel pipettes always search a tiprack top to bottom, left to right
            # - Eight channel pipettes will begin at the top if the primary nozzle is H1 and at the bottom if
            #   it is A1. The eight channel will always progress across the columns left to right.
            # - 96 Channel pipettes will begin in the corner opposite their primary/starting nozzle (if starting nozzle = A1, enter tiprack at H12)
            #   The 96 channel will then progress towards the opposite corner, either going up or down, left or right depending on configuration.

            if num_channels == 1:
                return _cluster_search_A1(num_nozzle_cols, num_nozzle_rows)
            elif num_channels == 8:
                if nozzle_map.starting_nozzle == "A1":
                    return _cluster_search_H1(num_nozzle_cols, num_nozzle_rows)
                elif nozzle_map.starting_nozzle == "H1":
                    return _cluster_search_A1(num_nozzle_cols, num_nozzle_rows)
            elif num_channels == 96:
                if nozzle_map.starting_nozzle == "A1":
                    return _cluster_search_H12(num_nozzle_cols, num_nozzle_rows)
                elif nozzle_map.starting_nozzle == "A12":
                    return _cluster_search_H1(num_nozzle_cols, num_nozzle_rows)
                elif nozzle_map.starting_nozzle == "H1":
                    return _cluster_search_A12(num_nozzle_cols, num_nozzle_rows)
                elif nozzle_map.starting_nozzle == "H12":
                    return _cluster_search_A1(num_nozzle_cols, num_nozzle_rows)
                else:
                    raise ValueError(
                        f"Nozzle {nozzle_map.starting_nozzle} is an invalid starting tip for automatic tip pickup."
                    )
            else:
                raise RuntimeError(
                    "Invalid number of channels for automatic tip tracking."
                )
        else:
            if columns and num_tips == len(columns[0]):  # Get next tips for 8-channel
                column_head = [column[0] for column in columns]
                starting_column_index = 0

                if starting_tip_name:
                    for idx, column in enumerate(columns):
                        if starting_tip_name in column:
                            if starting_tip_name not in column_head:
                                starting_column_index = idx + 1
                            else:
                                starting_column_index = idx

                for column in columns[starting_column_index:]:
                    if not any(
                        wells[well] == _TipRackWellState.USED for well in column
                    ):
                        return column[0]

            elif num_tips == len(wells.keys()):  # Get next tips for 96 channel
                if starting_tip_name and starting_tip_name != columns[0][0]:
                    return None

                if not any(
                    tip_state == _TipRackWellState.USED for tip_state in wells.values()
                ):
                    return next(iter(wells))

            else:  # Get next tips for single channel
                if starting_tip_name is not None:
                    wells = _drop_wells_before_starting_tip(wells, starting_tip_name)

                for well_name, tip_state in wells.items():
                    if tip_state == _TipRackWellState.CLEAN:
                        return well_name
        return None

    def has_clean_tip(self, labware_id: str, well_name: str) -> bool:
        """Get whether a well in a labware has a clean tip.

        Args:
            labware_id: The labware ID to check.
            well_name: The well name to check.

        Returns:
            True if the labware is a tip rack and the well has a clean tip,
            otherwise False.
        """
        tip_rack = self._state.tips_by_labware_id.get(labware_id)
        well_state = tip_rack.get(well_name) if tip_rack else None

        return well_state == _TipRackWellState.CLEAN

    def compute_tips_to_mark_as_used(
        self, labware_id: str, well_name: str, nozzle_map: NozzleMap
    ) -> list[str]:
        """Compute which tips a hypothetical tip pickup should mark as "used".

        Params:
            labware_id: The labware ID of the tip rack.
            well_name: The single target well of the tip pickup.
            nozzle_map: The nozzle configuration that the pipette will use for the pickup.

        Returns:
            The well names of all the tips that the operation will use.
        """
        columns = self._state.columns_by_labware_id.get(labware_id, [])
        return list(wells_covered_dense(nozzle_map, well_name, columns))


def _drop_wells_before_starting_tip(
    wells: _TipRackStateByWellName, starting_tip_name: str
) -> _TipRackStateByWellName:
    """Drop any wells that come before the starting tip and return the remaining ones after."""
    seen_starting_well = False
    remaining_wells: dict[str, _TipRackWellState] = {}
    for well_name, tip_state in wells.items():
        if well_name == starting_tip_name:
            seen_starting_well = True
        if seen_starting_well:
            remaining_wells[well_name] = tip_state
    return remaining_wells
